"""
PyTest configuration file for all tests
"""

import os
import json
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from flowapp import create_app
from flowapp import db as _db
from datetime import datetime
import flowapp.models
from flowapp.models.organization import Organization


TESTDB = "test_project.db"
TESTDB_PATH = "/tmp/{}".format(TESTDB)
TEST_DATABASE_URI = "sqlite:///" + TESTDB_PATH


class FieldMock:
    def __init__(self):
        self.data = None
        self.errors = []


class RuleMock:
    def __init__(self):
        self.source = None
        self.source_mask = None
        self.dest = None
        self.dest_mask = None


@pytest.fixture
def field():
    return FieldMock()


@pytest.fixture
def field_class():
    return FieldMock


@pytest.fixture
def rule():
    return RuleMock()


@pytest.fixture(scope="session")
def app(request):
    """
    Create a Flask app, and override settings, for the whole test session.
    """

    _app = create_app()

    _app.config.update(
        EXA_API="HTTP",
        EXA_API_URL="http://localhost:5000/",
        TESTING=True,
        SQLALCHEMY_DATABASE_URI=TEST_DATABASE_URI,
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        JWT_SECRET="testing",
        API_KEY="testkey",
        SECRET_KEY="testkeysession",
        LOCAL_USER_UUID="jiri.vrany@cesnet.cz",
        LOCAL_AUTH=True,
        ALLOWED_COMMUNITIES=[1, 2, 3],
        WTF_CSRF_ENABLED=False,
    )

    print("\n----- CREATE FLASK APPLICATION\n")
    context = _app.app_context()
    context.push()
    yield _app
    print("\n----- CREATE FLASK APPLICATION CONTEXT\n")

    context.pop()
    print("\n----- RELEASE FLASK APPLICATION CONTEXT\n")


@pytest.fixture(scope="session")
def client(app, request):
    """
    Get the test_client from the app, for the whole test session.
    """
    print("\n----- CREATE FLASK TEST CLIENT\n")
    return app.test_client()


@pytest.fixture(scope="session")
def db(app, request):
    """
    Create entire database for every test.
    """
    engine = create_engine(app.config["SQLALCHEMY_DATABASE_URI"], echo=True)
    sessionmaker(bind=engine)
    print("\n----- CREATE TEST DB CONNECTION POOL\n")
    if os.path.exists(TESTDB_PATH):
        os.unlink(TESTDB_PATH)

    with app.app_context():
        _db.init_app(app)
        print("#: cleaning database")
        _db.reflect()
        _db.drop_all()
        print("#: creating tables")
        _db.create_all()

        users = [
            {"name": "jiri.vrany@cesnet.cz", "role_id": 3, "org_id": 1},
            {"name": "petr.adamec@cesnet.cz", "role_id": 3, "org_id": 1},
        ]
        print("#: inserting users")
        flowapp.models.insert_users(users)

        org = _db.session.query(Organization).filter_by(id=1).first()
        # Update the organization address range to include our test networks
        org.arange = "147.230.0.0/16\n2001:718:1c01::/48\n192.168.0.0/16\n10.0.0.0/8"
        _db.session.commit()
        print("\n----- UPDATED TEST ORG 1 \n", org)

    def teardown():
        _db.session.commit()
        _db.drop_all()
        os.unlink(TESTDB_PATH)

    request.addfinalizer(teardown)
    return _db


@pytest.fixture(scope="session")
def jwt_token(client, app, db, request):
    """
    Get the test_client from the app, for the whole test session.
    """
    mkey = "testkey"

    with app.app_context():
        model = flowapp.models.ApiKey(machine="127.0.0.1", key=mkey, user_id=1, org_id=1)
        db.session.add(model)
        db.session.commit()

    print("\n----- GET JWT TEST TOKEN\n")
    url = "/api/v3/auth"
    headers = {"x-api-key": mkey}
    token = client.get(url, headers=headers)
    data = json.loads(token.data)
    return data["token"]


@pytest.fixture(scope="session")
def machine_api_token(client, app, db, request):
    """
    Get the test_client from the app, for the whole test session.
    """
    mkey = "machinetestkey"

    with app.app_context():
        model = flowapp.models.MachineApiKey(machine="127.0.0.1", key=mkey, user_id=1, org_id=1)
        db.session.add(model)
        db.session.commit()

    print("\n----- GET MACHINE API KEY TEST TOKEN\n")
    url = "/api/v3/auth"
    headers = {"x-api-key": mkey}
    token = client.get(url, headers=headers)
    data = json.loads(token.data)
    return data["token"]


@pytest.fixture(scope="session")
def expired_auth_token(client, app, db, request):
    """
    Get the test_client from the app, for the whole test session.
    """
    test_key = "expired_test_key"
    expired_date = datetime.strptime("2019-01-01", "%Y-%m-%d")
    with app.app_context():
        model = flowapp.models.ApiKey(machine="127.0.0.1", key=test_key, user_id=1, expires=expired_date, org_id=1)
        db.session.add(model)
        db.session.commit()

    return test_key


@pytest.fixture(scope="session")
def readonly_jwt_token(client, app, db, request):
    """
    Get the test_client from the app, for the whole test session.
    """
    readonly_key = "readonly-testkey"
    with app.app_context():
        model = flowapp.models.ApiKey(machine="127.0.0.1", key=readonly_key, user_id=1, readonly=True, org_id=1)
        db.session.add(model)
        db.session.commit()

    print("\n----- GET JWT TEST TOKEN\n")
    url = "/api/v3/auth"
    headers = {"x-api-key": readonly_key}
    token = client.get(url, headers=headers)
    data = json.loads(token.data)
    return data["token"]


@pytest.fixture(scope="session")
def auth_client(client):
    """
    Get the test_client from the app, for the whole test session.
    """
    print("\n----- CREATE AUTHENTICATED FLASK TEST CLIENT\n")
    client.get("/local-login")
    return client


@pytest.fixture(autouse=True)
def reset_org_limits(db, app):
    """
    Automatically reset organization limits after each test that modifies them.
    """
    yield  # Allow test execution

    with app.app_context():
        org = db.session.query(Organization).filter_by(id=1).first()
        if org:
            org.limit_flowspec4 = 0
            org.limit_flowspec6 = 0
            org.limit_rtbh = 0
            db.session.commit()



import re

from flowapp.constants import MAX_PORT

NUMBER = re.compile(r"^\d+$", re.IGNORECASE)
RANGE = re.compile(r"^(\d+)-(\d+)$", re.IGNORECASE)
NOTRAN = re.compile(r"^>=(\d+)&<=(\d+)$", re.IGNORECASE)
GREATER = re.compile(r">[=]?(\d+)$", re.IGNORECASE)
LOWER = re.compile(r"<[=]?(\d+)$", re.IGNORECASE)


def translate_sequence(sequence, max_val=MAX_PORT):
    """
    translate command sequence sepparated by ; to ExaBGP command format
    @param string sequence
    @param integer max_val
    @return string ExaBgp rule string
    """
    result = [to_exabgp_string(item, max_val) for item in sequence.split(";") if item]
    result = " ".join(result)
    return "[{}]".format(result)


def check_limit(value, max_value, min_value=0):
    """
    test if the value is within valid range (min_value to max_value inclusive)
    raise exception otherwise
    """
    value = int(value)
    if value > max_value:
        raise ValueError("Invalid value number: {} is too big. Max is {}.".format(value, max_value))
    if value < min_value:
        raise ValueError("Invalid value number: {} is too small. Min is {}.".format(value, min_value))
    return value


def to_exabgp_string(value_string, max_val):
    """
    Translate form string to flowspec value or packet size rule
    @param string value_string
    @param integer max_val
    @return string ExaBgp rule string
    @raises ValueError

    x (integer)  to =x
    x-y  to >=x&<=y
    >=x&<=y to >=x&<=y
    >x to >=x&<=MAX
    >=x to >=x&<=MAX
    <x to >=0&<=x
    <=x to >=0&<=x
    """
    # simple number
    if NUMBER.match(value_string):
        return "={}".format(check_limit(value_string, max_val))
    elif RANGE.match(value_string):
        m = RANGE.match(value_string)
        start = check_limit(m.group(1), max_val)
        end = check_limit(m.group(2), max_val)
        if start > end:
            raise ValueError("Invalid range: start value cannot be greater than end value")
        return ">={}&<={}".format(start, end)
    elif NOTRAN.match(value_string):
        m = NOTRAN.match(value_string)
        start = check_limit(m.group(1), max_val)
        end = check_limit(m.group(2), max_val)
        if start > end:
            raise ValueError("Invalid range: start value cannot be greater than end value")
        return ">={}&<={}".format(start, end)
    elif GREATER.match(value_string):
        m = GREATER.match(value_string)
        return ">={}&<={}".format(check_limit(m.group(1), max_val), max_val)
    elif LOWER.match(value_string):
        m = LOWER.match(value_string)
        # Even for lower bound expressions, validate that the value itself is within range
        end = check_limit(m.group(1), max_val)
        return ">=0&<={}".format(end)
    else:
        raise ValueError("string {} can not be converted".format(value_string))


def filter_rules_action(user_actions, rules):
    """
    Divide the list of rules by user_actions to editable and viewonly subsets
    :param user_actions: list of actions allowed for normal user
    :param rules: list of rules to be filtered
    :return: editable, viewonly lists
    """
    editable = []
    viewonly = []
    for rule in rules:
        try:
            if rule.action_id in user_actions:
                editable.append(rule)
            else:
                viewonly.append(rule)
        except AttributeError:
            editable.append(rule)  # If rule has no action_id, treat it as editable

    return editable, viewonly



# Test for api authorization
import json


def test_token(client, jwt_token):
    """
    test that token authorization works
    """
    req = client.get("/api/v3/test_token", headers={"x-access-token": jwt_token})

    assert req.status_code == 200


def test_machine_token(client, machine_api_token):
    """
    test that token authorization works
    """
    req = client.get("/api/v3/test_token", headers={"x-access-token": machine_api_token})

    assert req.status_code == 200


def test_expired_token(client, expired_auth_token):
    """
    test that expired token authorization return 401
    """
    req = client.get("/api/v3/auth", headers={"x-api-key": expired_auth_token})

    assert req.status_code == 401


def test_withnout_token(client):
    """
    test that without token authorization return 401
    """
    req = client.get("/api/v3/test_token")

    assert req.status_code == 401


def test_readonly_token(client, readonly_jwt_token):
    """
    test that readonly flag is set correctly
    """
    req = client.get("/api/v3/test_token", headers={"x-access-token": readonly_jwt_token})

    assert req.status_code == 200
    data = json.loads(req.data)
    assert data["readonly"]


def test_readonly_token_ipv4_create(client, db, readonly_jwt_token):
    """
    test that readonly token can't create ipv4 rule
    """
    headers = {"x-access-token": readonly_jwt_token}

    req = client.post(
        "/api/v3/rules/ipv4",
        headers=headers,
        json={
            "action": 2,
            "protocol": "tcp",
            "source": "147.230.17.117",
            "source_mask": 32,
            "source_port": "",
            "expires": "1444913400",
        },
    )

    assert req.status_code == 403



"""
IPv6 rule form for the flowapp application.
"""

from wtforms import StringField, IntegerField, SelectField
from wtforms.validators import Optional, NumberRange, DataRequired

from ...constants import IPV6_NEXT_HEADER
from ...validators import IPv6Address
from .base import IPForm


class IPv6Form(IPForm):
    """
    IPv6 form object
    """

    def __init__(self, *args, **kwargs):
        super(IPv6Form, self).__init__(*args, **kwargs)
        self.net_ranges = None

    zero_address = "::"
    source = StringField(
        "Source address",
        validators=[Optional(), IPv6Address(message="provide valid IPv6 adress")],
    )

    source_mask = IntegerField(
        "Source prefix length (bits)",
        validators=[
            Optional(),
            NumberRange(min=0, max=128, message="invalid prefix value (0-128)"),
        ],
    )

    dest = StringField(
        "Destination address",
        validators=[Optional(), IPv6Address(message="provide valid IPv6 adress")],
    )

    dest_mask = IntegerField(
        "Destination prefix length (bits)",
        validators=[
            Optional(),
            NumberRange(min=0, max=128, message="invalid prefix value (0-128)"),
        ],
    )

    next_header = SelectField(
        "Next Header",
        choices=[(pr, pr.upper()) for pr in IPV6_NEXT_HEADER.keys()],
        validators=[DataRequired()],
    )

    def validate_ipv_specific(self):
        """
        validate next header and flags, set error message if validation fails
        :return: boolean validation result
        """
        if len(self.flags.data) > 0 and self.next_header.data != "tcp":
            self.flags.errors.append("Can not set TCP flags for next-header {} !".format(self.next_header.data.upper()))
            return False

        return True



import pytest
from flowapp.services.whitelist_common import (
    Relation,
    check_rule_against_whitelists,
    check_whitelist_against_rules,
    check_whitelist_to_rule_relation,
    subtract_network,
    clear_network_cache,
    _is_same_ip_version,  # New helper function
)


# Tests for core function that checks network relations
@pytest.mark.parametrize(
    "rule,whitelist,expected",
    [
        # IPv4 test cases
        ("192.168.1.0/24", "192.168.0.0/16", Relation.SUPERNET),  # Whitelist is supernet
        ("192.168.1.0/24", "192.168.1.0/24", Relation.EQUAL),  # Equal networks
        ("192.168.1.0/24", "192.168.1.128/25", Relation.SUBNET),  # Whitelist is subnet
        ("192.168.1.128/25", "192.168.1.0/24", Relation.SUPERNET),  # Whitelist is supernet
        ("10.0.0.0/8", "192.168.1.0/24", Relation.DIFFERENT),  # Different networks
        # IPv6 test cases
        ("2001:db8::/32", "2001:db8::/32", Relation.EQUAL),  # Equal networks
        ("2001:db8:1::/48", "2001:db8::/32", Relation.SUPERNET),  # Whitelist is supernet
        ("2001:db8::/32", "2001:db8:1::/48", Relation.SUBNET),  # Whitelist is subnet
        ("2001:db8::/32", "2002:db8::/32", Relation.DIFFERENT),  # Different networks
        ("2001:db8:1:2::/64", "2001:db8::/32", Relation.SUPERNET),  # Whitelist is supernet
    ],
)
def test_check_whitelist_to_rule_relation(rule, whitelist, expected):
    clear_network_cache()
    assert check_whitelist_to_rule_relation(rule, whitelist) == expected


@pytest.mark.parametrize(
    "target,whitelist,expected",
    [
        # Basic IPv4 cases
        (
            "192.168.1.0/24",
            "192.168.1.128/25",
            ["192.168.1.0/25"],
        ),  # One remaining subnet
        (
            "192.168.1.0/24",
            "192.168.1.64/26",
            ["192.168.1.0/26", "192.168.1.128/25"],
        ),  # Two remaining subnets
        (
            "192.168.1.0/24",
            "192.168.1.0/24",
            ["192.168.1.0/24"],
        ),  # Equal networks - return original network
        (
            "192.168.1.0/24",
            "192.168.2.0/24",
            ["192.168.1.0/24"],
        ),  # No overlap
    ],
)
def test_subtract_network(target, whitelist, expected):
    clear_network_cache()
    result = subtract_network(target, whitelist)
    assert sorted(result) == sorted(expected)


def test_check_rule_against_whitelists():
    clear_network_cache()

    rule = "192.168.1.0/24"
    whitelists = [
        "192.168.0.0/16",  # SUPERNET
        "172.16.0.0/12",  # DIFFERENT
        "192.168.1.0/24",  # EQUAL
        "192.168.1.128/25",  # SUBNET
    ]

    results = check_rule_against_whitelists(rule, whitelists)

    # Should return list of tuples (rule, whitelist, relation) for non-DIFFERENT relations
    expected = [
        (rule, "192.168.0.0/16", Relation.SUPERNET),
        (rule, "192.168.1.0/24", Relation.EQUAL),
        (rule, "192.168.1.128/25", Relation.SUBNET),
    ]

    assert len(results) == 3
    for result, exp in zip(sorted(results), sorted(expected)):
        assert result == exp


def test_check_whitelist_against_rules():
    clear_network_cache()

    whitelist = "192.168.1.128/25"
    rules = [
        "192.168.0.0/16",  # Whitelist is SUBNET of this larger network
        "172.16.0.0/12",  # DIFFERENT
        "192.168.1.128/25",  # EQUAL
        "192.168.1.0/24",  # Whitelist is SUBNET of this network
    ]

    results = check_whitelist_against_rules(rules, whitelist)

    # Should return list of tuples (rule, whitelist, relation) for non-DIFFERENT relations
    expected = [
        ("192.168.0.0/16", whitelist, Relation.SUBNET),
        ("192.168.1.128/25", whitelist, Relation.EQUAL),
        ("192.168.1.0/24", whitelist, Relation.SUBNET),
    ]

    assert len(results) == 3
    for result, exp in zip(sorted(results), sorted(expected)):
        assert result == exp


# Tests for edge cases and error handling
def test_host_addresses():
    clear_network_cache()
    # Test with host addresses (no explicit subnet mask)
    assert check_whitelist_to_rule_relation("192.168.1.1", "192.168.1.0/24") == Relation.SUPERNET
    assert check_whitelist_to_rule_relation("192.168.1.1", "192.168.2.0/24") == Relation.DIFFERENT


def test_single_ip_as_network():
    clear_network_cache()
    # Test with /32 (single IPv4 address as network)
    assert check_whitelist_to_rule_relation("192.168.1.1/32", "192.168.1.0/24") == Relation.SUPERNET
    # Test with /128 (single IPv6 address as network)
    assert check_whitelist_to_rule_relation("2001:db8::1/128", "2001:db8::/32") == Relation.SUPERNET


def test_is_same_ip_version():
    """Test the new helper function to check if two addresses are of the same IP version"""
    # IPv4 cases
    assert _is_same_ip_version("192.168.1.0/24", "10.0.0.0/8") is True
    assert _is_same_ip_version("192.168.1.1", "10.0.0.1") is True

    # IPv6 cases
    assert _is_same_ip_version("2001:db8::/32", "fe80::/64") is True
    assert _is_same_ip_version("2001:db8::1", "fe80::1") is True

    # Mixed cases
    assert _is_same_ip_version("192.168.1.0/24", "2001:db8::/32") is False
    assert _is_same_ip_version("192.168.1.1", "fe80::1") is False


def test_check_whitelist_to_rule_relation_mixed_versions():
    """Test the check_whitelist_to_rule_relation function with mixed IP versions"""
    clear_network_cache()

    # IPv4 rule with IPv6 whitelist
    assert check_whitelist_to_rule_relation("192.168.1.0/24", "2001:db8::/32") == Relation.DIFFERENT

    # IPv6 rule with IPv4 whitelist
    assert check_whitelist_to_rule_relation("2001:db8::/32", "192.168.1.0/24") == Relation.DIFFERENT

    # Invalid IP addresses should return DIFFERENT
    assert check_whitelist_to_rule_relation("invalid", "192.168.1.0/24") == Relation.DIFFERENT
    assert check_whitelist_to_rule_relation("192.168.1.0/24", "invalid") == Relation.DIFFERENT


def test_subtract_network_mixed_versions():
    """Test the subtract_network function with mixed IP versions"""
    clear_network_cache()

    # IPv4 target with IPv6 whitelist - should return original target
    result = subtract_network("192.168.1.0/24", "2001:db8::/32")
    assert result == ["192.168.1.0/24"]

    # IPv6 target with IPv4 whitelist - should return original target
    result = subtract_network("2001:db8::/32", "192.168.1.0/24")
    assert result == ["2001:db8::/32"]

    # Invalid addresses - should return original target
    result = subtract_network("invalid", "192.168.1.0/24")
    assert result == ["invalid"]
    result = subtract_network("192.168.1.0/24", "invalid")
    assert result == ["192.168.1.0/24"]


def test_check_rule_against_whitelists_mixed_versions():
    """Test the check_rule_against_whitelists function with mixed IP versions"""
    clear_network_cache()

    # IPv4 rule against mixed whitelist
    rule = "192.168.1.0/24"
    whitelists = [
        "192.168.0.0/16",  # IPv4 - should match
        "2001:db8::/32",  # IPv6 - should not match
        "10.0.0.0/8",  # IPv4 - should not match (different network)
    ]

    results = check_rule_against_whitelists(rule, whitelists)
    assert len(results) == 1
    assert results[0][0] == rule
    assert results[0][1] == "192.168.0.0/16"
    assert results[0][2] == Relation.SUPERNET

    # IPv6 rule against mixed whitelist
    rule = "2001:db8:1::/48"
    whitelists = [
        "192.168.0.0/16",  # IPv4 - should not match
        "2001:db8::/32",  # IPv6 - should match
        "2002:db8::/32",  # IPv6 - should not match (different network)
    ]

    results = check_rule_against_whitelists(rule, whitelists)
    assert len(results) == 1
    assert results[0][0] == rule
    assert results[0][1] == "2001:db8::/32"
    assert results[0][2] == Relation.SUPERNET


def test_check_whitelist_against_rules_mixed_versions():
    """Test the check_whitelist_against_rules function with mixed IP versions"""
    clear_network_cache()

    # IPv4 whitelist against mixed rules
    whitelist = "192.168.1.0/24"
    rules = [
        "192.168.0.0/16",  # IPv4 - should match
        "2001:db8::/32",  # IPv6 - should not match
        "10.0.0.0/8",  # IPv4 - should not match (different network)
    ]

    results = check_whitelist_against_rules(rules, whitelist)
    assert len(results) == 1
    assert results[0][0] == "192.168.0.0/16"
    assert results[0][1] == whitelist
    assert results[0][2] == Relation.SUBNET

    # IPv6 whitelist against mixed rules
    whitelist = "2001:db8:1::/48"
    rules = [
        "192.168.0.0/16",  # IPv4 - should not match
        "2001:db8::/32",  # IPv6 - should match
        "2002:db8::/32",  # IPv6 - should not match (different network)
    ]

    results = check_whitelist_against_rules(rules, whitelist)
    assert len(results) == 1
    assert results[0][0] == "2001:db8::/32"
    assert results[0][1] == whitelist
    assert results[0][2] == Relation.SUBNET


if __name__ == "__main__":
    pytest.main(["-v"])



"""
Organization form for the flowapp application.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, TextAreaField
from wtforms.validators import Optional, Length, NumberRange

from ..validators import NetRangeString


class OrganizationForm(FlaskForm):
    """
    Organization form object
    used in Admin
    """

    name = StringField("Organization name", validators=[Optional(), Length(max=150)])

    limit_flowspec4 = IntegerField(
        "Maximum number of IPv4 rules, 0 for unlimited",
        validators=[
            Optional(),
            NumberRange(min=0, max=1000, message="invalid mask value (0-1000)"),
        ],
    )

    limit_flowspec6 = IntegerField(
        "Maximum number of IPv6 rules, 0 for unlimited",
        validators=[
            Optional(),
            NumberRange(min=0, max=1000, message="invalid mask value (0-1000)"),
        ],
    )

    limit_rtbh = IntegerField(
        "Maximum number of RTBH rules, 0 for unlimited",
        validators=[
            Optional(),
            NumberRange(min=0, max=1000, message="invalid mask value (0-1000)"),
        ],
    )

    arange = TextAreaField(
        "Organization Adress Range - one range per row",
        validators=[Optional(), NetRangeString()],
    )



import csv
from io import StringIO
from datetime import datetime, timedelta
import secrets

from sqlalchemy import func, text
from flask import Blueprint, render_template, redirect, flash, request, session, url_for, current_app
import sqlalchemy
from sqlalchemy.exc import IntegrityError, OperationalError

from ..forms import ASPathForm, BulkUserForm, MachineApiKeyForm, UserForm, ActionForm, OrganizationForm, CommunityForm
from ..models import (
    ASPath,
    ApiKey,
    MachineApiKey,
    User,
    Action,
    Organization,
    Role,
    insert_user,
    get_existing_action,
    Community,
    get_existing_community,
    Log,
    Flowspec4,
    Flowspec6,
    RTBH,
)
from ..auth import auth_required, admin_required
from flowapp import db

admin = Blueprint("admin", __name__, template_folder="templates")


@admin.route("/log", methods=["GET"], defaults={"page": 1})
@admin.route("/log/<int:page>", methods=["GET"])
@auth_required
@admin_required
def log(page):
    """
    Displays logs for last two days
    """
    per_page = 20
    now = datetime.now()
    week_ago = now - timedelta(weeks=1)
    logs = (
        Log.query.order_by(Log.time.desc())
        .filter(Log.time > week_ago)
        .paginate(page=page, per_page=per_page, max_per_page=None, error_out=False)
    )
    return render_template("pages/logs.html", logs=logs)


@admin.route("/machine_keys", methods=["GET"])
@auth_required
@admin_required
def machine_keys():
    """
    Display all machine keys, from all admins
    """
    keys = db.session.query(MachineApiKey).all()

    return render_template("pages/machine_api_key.html", keys=keys)


@admin.route("/add_machine_key", methods=["GET", "POST"])
@auth_required
@admin_required
def add_machine_key():
    """
    Add new MachnieApiKey
    :return: form or redirect to list of keys
    """
    generated = secrets.token_hex(24)
    form = MachineApiKeyForm(request.form, key=generated)
    form.user.choices = [(g.id, f"{g.name} ({g.uuid})") for g in db.session.query(User).order_by("name")]

    if request.method == "POST" and form.validate():
        target_user = db.session.get(User, form.user.data)
        target_org = target_user.organization.first() if target_user else None
        current_user = session.get("user_name")
        curent_email = session.get("user_uuid")
        comment = f"created by: {current_user}/{curent_email}, comment: {form.comment.data}"
        model = MachineApiKey(
            machine=form.machine.data,
            key=form.key.data,
            expires=form.expires.data,
            readonly=form.readonly.data,
            comment=comment,
            user_id=target_user.id,
            org_id=target_org.id,
        )

        db.session.add(model)
        db.session.commit()
        flash("NewKey saved", "alert-success")

        return redirect(url_for("admin.machine_keys"))
    else:
        for field, errors in form.errors.items():
            for error in errors:
                current_app.logger.debug("Error in the %s field - %s" % (getattr(form, field).label.text, error))

    return render_template("forms/machine_api_key.html", form=form, generated_key=generated)


@admin.route("/delete_machine_key/<int:key_id>", methods=["GET"])
@auth_required
@admin_required
def delete_machine_key(key_id):
    """
    Delete api_key and machine
    :param key_id: integer
    """
    model = db.session.get(MachineApiKey, key_id)
    # delete from db
    db.session.delete(model)
    db.session.commit()
    flash("Key deleted", "alert-success")

    return redirect(url_for("admin.machine_keys"))


@admin.route("/user", methods=["GET", "POST"])
@auth_required
@admin_required
def user():
    form = UserForm(request.form)
    form.role_ids.choices = [(g.id, g.name) for g in db.session.query(Role).order_by("name")]
    form.org_ids.choices = [(g.id, g.name) for g in db.session.query(Organization).order_by("name")]

    if request.method == "POST" and form.validate():
        # test if user is unique
        exist = db.session.query(User).filter_by(uuid=form.uuid.data).first()
        if not exist:
            insert_user(
                uuid=form.uuid.data,
                name=form.name.data,
                phone=form.phone.data,
                email=form.email.data,
                comment=form.comment.data,
                role_ids=form.role_ids.data,
                org_ids=form.org_ids.data,
            )
            flash("User saved")
            return redirect(url_for("admin.users"))
        else:
            flash("User {} already exists".format(form.email.data), "alert-danger")

    action_url = url_for("admin.user")
    return render_template(
        "forms/simple_form.html",
        title="Add new user to Flowspec",
        form=form,
        action_url=action_url,
    )


@admin.route("/user/edit/<int:user_id>", methods=["GET", "POST"])
@auth_required
@admin_required
def edit_user(user_id):
    user = db.session.get(User, user_id)
    form = UserForm(request.form, obj=user)
    form.role_ids.choices = [(g.id, g.name) for g in db.session.query(Role).order_by("name")]
    form.org_ids.choices = [(g.id, g.name) for g in db.session.query(Organization).order_by("name")]

    if request.method == "POST" and form.validate():
        user.update(form)
        return redirect(url_for("admin.users"))

    form.role_ids.data = [role.id for role in user.role]
    form.org_ids.data = [org.id for org in user.organization]
    action_url = url_for("admin.edit_user", user_id=user_id)

    return render_template(
        "forms/simple_form.html",
        title=f"Editing {user.email}",
        form=form,
        action_url=action_url,
    )


@admin.route("/user/delete/<int:user_id>", methods=["GET"])
@auth_required
@admin_required
def delete_user(user_id):
    user = db.session.get(User, user_id)
    if not user:
        flash("User not found.", "alert-danger")
        return redirect(url_for("admin.users"))

    username = user.email
    db.session.delete(user)

    message = f"User {username} deleted"
    alert_type = "alert-success"

    try:
        db.session.commit()
    except IntegrityError:
        db.session.rollback()  # Rollback on IntegrityError
        message = f"User {username} owns some rules, cannot be deleted! Delete rules first."
        alert_type = "alert-danger"
    except OperationalError:
        db.session.rollback()  # Rollback on OperationalError
        message = f"User {username} owns some rules, cannot be deleted! Delete rules first."
        alert_type = "alert-danger"

    flash(message, alert_type)
    return redirect(url_for("admin.users"))


@admin.route("/users")
@auth_required
@admin_required
def users():
    users = User.query.all()
    return render_template("pages/users.html", users=users)


@admin.route("/bulk-import-users", methods=["GET"])
@auth_required
@admin_required
def bulk_import_users():
    form = BulkUserForm(request.form)
    orgs = db.session.execute(db.select(Organization).order_by(Organization.name)).scalars()
    return render_template("forms/bulk_user_form.html", form=form, orgs=orgs)


@admin.route("/bulk-import-users", methods=["POST"])
@auth_required
@admin_required
def bulk_import_users_save():
    form = BulkUserForm(request.form)
    roles = [role.id for role in db.session.query(Role).all()]
    orgs = [org.id for org in db.session.query(Organization).all()]
    uuids = [user.uuid for user in db.session.query(User).all()]
    form.roles = roles
    form.organizations = orgs
    form.uuids = uuids

    if request.method == "POST" and form.validate():
        # Get CSV data from textarea
        csv_data = form.users.data
        # Parse CSV data
        csv_reader = csv.DictReader(StringIO(csv_data), delimiter=",")
        errored = False
        for row in csv_reader:
            try:
                # Extract and prepare data
                uuid = row["uuid-eppn"]
                name = row["name"]
                phone = row["telefon"]
                email = row["email"]

                # Convert role and organization fields to lists of integers
                role_ids = [int(row["role"])]  # role_id should be a list
                org_ids = [int(row["organizace"])]  # org_id should be a list
                notice = row["poznamka"]

                # Insert user
                insert_user(
                    uuid=uuid, role_ids=role_ids, org_ids=org_ids, name=name, phone=phone, email=email, comment=notice
                )
            except KeyError as e:
                errored = True
                # Handle missing fields or other errors in the CSV
                flash(f"Missing field in CSV: {e}", "alert-danger")
            except ValueError as e:
                errored = True
                # Handle conversion issues (like invalid int for role/org)
                flash(f"Data conversion error: {e}", "alert-danger")
            except sqlalchemy.exc.IntegrityError as e:
                errored = True
                db.session.rollback()
                flash(f"SQL Integrity error: {e}", "alert-danger")

        if not errored:
            return redirect(url_for("admin.users"))

    return render_template("forms/bulk_user_form.html", form=form)


@admin.route("/organizations")
@auth_required
@admin_required
def organizations():
    # Query all organizations and eager load RTBH relationships
    orgs = db.session.query(Organization).options(db.joinedload(Organization.rtbh)).all()

    # Get RTBH counts with rstate_id=1 for all organizations in one query
    rtbh_counts_query = (
        db.session.query(RTBH.org_id, func.count(RTBH.id)).filter(RTBH.rstate_id == 1).group_by(RTBH.org_id).all()
    )

    flowspec4_count_query = (
        db.session.query(Flowspec4.org_id, func.count(Flowspec4.id))
        .filter(Flowspec4.rstate_id == 1)
        .group_by(Flowspec4.org_id)
        .all()
    )

    flowspec6_count_query = (
        db.session.query(Flowspec6.org_id, func.count(Flowspec6.id))
        .filter(Flowspec6.rstate_id == 1)
        .group_by(Flowspec6.org_id)
        .all()
    )

    flowspec4_all_count = db.session.query(Flowspec4).filter(Flowspec4.rstate_id == 1).count()
    flowspec6_all_count = db.session.query(Flowspec6).filter(Flowspec6.rstate_id == 1).count()
    rtbh_all_count = db.session.query(RTBH).filter(RTBH.rstate_id == 1).count()

    # Convert query result to a dictionary {org_id: count}
    rtbh_counts = {org_id: count for org_id, count in rtbh_counts_query}
    flowspec4_counts = {org_id: count for org_id, count in flowspec4_count_query}
    flowspec6_counts = {org_id: count for org_id, count in flowspec6_count_query}

    return render_template(
        "pages/orgs.html",
        orgs=orgs,
        rtbh_counts=rtbh_counts,
        flowspec4_counts=flowspec4_counts,
        flowspec6_counts=flowspec6_counts,
        rtbh_all_count=rtbh_all_count,
        flowspec4_all_count=flowspec4_all_count,
        flowspec6_all_count=flowspec6_all_count,
        flowspec4_limit=current_app.config.get("FLOWSPEC4_MAX_RULES", 9000),
        flowspec6_limit=current_app.config.get("FLOWSPEC6_MAX_RULES", 9000),
        rtbh_limit=current_app.config.get("RTBH_MAX_RULES", 100000),
    )


@admin.route("/organization", methods=["GET", "POST"])
@auth_required
@admin_required
def organization():
    form = OrganizationForm(request.form)

    if request.method == "POST" and form.validate():
        # test if user is unique
        exist = db.session.query(Organization).filter_by(name=form.name.data).first()
        if not exist:
            org = Organization(
                name=form.name.data,
                arange=form.arange.data,
                limit_flowspec4=form.limit_flowspec4.data,
                limit_flowspec6=form.limit_flowspec6.data,
                limit_rtbh=form.limit_rtbh.data,
            )
            db.session.add(org)
            db.session.commit()
            flash("Organization saved")
            return redirect(url_for("admin.organizations"))
        else:
            flash("Organization {} already exists".format(form.name.data), "alert-danger")

    action_url = url_for("admin.organization")
    return render_template(
        "forms/org.html",
        title="Add new organization to Flowspec",
        form=form,
        action_url=action_url,
    )


@admin.route("/organization/edit/<int:org_id>", methods=["GET", "POST"])
@auth_required
@admin_required
def edit_organization(org_id):
    org = db.session.get(Organization, org_id)
    form = OrganizationForm(request.form, obj=org)

    if request.method == "POST" and form.validate():
        form.populate_obj(org)
        db.session.commit()
        flash("Organization updated", "alert-success")
        return redirect(url_for("admin.organizations"))

    action_url = url_for("admin.edit_organization", org_id=org.id)
    return render_template(
        "forms/org.html",
        title="Editing {}".format(org.name),
        form=form,
        action_url=action_url,
    )


@admin.route("/organization/delete/<int:org_id>", methods=["GET"])
@auth_required
@admin_required
def delete_organization(org_id):
    org = db.session.get(Organization, org_id)
    aname = org.name
    db.session.delete(org)
    message = "Organization {} deleted".format(aname)
    alert_type = "alert-success"
    try:
        db.session.commit()
    except IntegrityError:
        message = "Organization {} has some users, can not be deleted!".format(aname)
        alert_type = "alert-danger"

    flash(message, alert_type)
    db.session.commit()

    return redirect(url_for("admin.organizations"))


@admin.route("/as-paths")
@auth_required
@admin_required
def as_paths():
    mpaths = db.session.query(ASPath).all()
    return render_template("pages/as_paths.html", paths=mpaths)


@admin.route("/as-path", methods=["GET", "POST"])
@auth_required
@admin_required
def as_path():
    form = ASPathForm(request.form)

    if request.method == "POST" and form.validate():
        # test if user is unique
        exist = db.session.query(ASPath).filter_by(prefix=form.prefix.data).first()
        if not exist:
            pth = ASPath(prefix=form.prefix.data, as_path=form.as_path.data)
            db.session.add(pth)
            db.session.commit()
            flash("AS-path saved")
            return redirect(url_for("admin.as_paths"))
        else:
            flash("Prefix {} already taken".format(form.prefix.data), "alert-danger")

    action_url = url_for("admin.as_path")
    return render_template(
        "forms/simple_form.html",
        title="Add new AS-path to Flowspec",
        form=form,
        action_url=action_url,
    )


@admin.route("/as-path/edit/<int:path_id>", methods=["GET", "POST"])
@auth_required
@admin_required
def edit_as_path(path_id):
    pth = db.session.get(ASPath, path_id)
    form = ASPathForm(request.form, obj=pth)

    if request.method == "POST" and form.validate():
        form.populate_obj(pth)
        db.session.commit()
        flash("AS-path updated")
        return redirect(url_for("admin.as_paths"))

    action_url = url_for("admin.edit_as_path", path_id=pth.id)
    return render_template(
        "forms/simple_form.html",
        title="Editing {}".format(pth.prefix),
        form=form,
        action_url=action_url,
    )


@admin.route("/as-path/delete/<int:path_id>", methods=["GET"])
@auth_required
@admin_required
def delete_as_path(path_id):
    pth = db.session.get(ASPath, path_id)
    db.session.delete(pth)
    message = f"AS path {pth.prefix} : {pth.as_path} deleted"
    alert_type = "alert-success"

    flash(message, alert_type)
    db.session.commit()

    return redirect(url_for("admin.as_paths"))


@admin.route("/actions")
@auth_required
@admin_required
def actions():
    actions = db.session.query(Action).all()
    return render_template("pages/actions.html", actions=actions)


@admin.route("/action", methods=["GET", "POST"])
@auth_required
@admin_required
def action():
    form = ActionForm(request.form)

    if request.method == "POST" and form.validate():
        # test if Acttion is unique
        exist = get_existing_action(form.name.data, form.command.data)
        if not exist:
            action = Action(
                name=form.name.data,
                command=form.command.data,
                description=form.description.data,
                role_id=form.role_id.data,
            )
            db.session.add(action)
            db.session.commit()
            flash("Action saved", "alert-success")
            return redirect(url_for("admin.actions"))
        else:
            flash(
                "Action with name {} or command {} already exists".format(form.name.data, form.command.data),
                "alert-danger",
            )

    action_url = url_for("admin.action")
    return render_template(
        "forms/simple_form.html",
        title="Add new action to Flowspec",
        form=form,
        action_url=action_url,
    )


@admin.route("/action/edit/<int:action_id>", methods=["GET", "POST"])
@auth_required
@admin_required
def edit_action(action_id):
    action = db.session.get(Action, action_id)
    form = ActionForm(request.form, obj=action)
    if request.method == "POST" and form.validate():
        form.populate_obj(action)
        db.session.commit()
        flash("Action updated")
        return redirect(url_for("admin.actions"))

    action_url = url_for("admin.edit_action", action_id=action.id)
    return render_template(
        "forms/simple_form.html",
        title="Editing {}".format(action.name),
        form=form,
        action_url=action_url,
    )


@admin.route("/action/delete/<int:action_id>", methods=["GET"])
@auth_required
@admin_required
def delete_action(action_id):
    action = db.session.get(Action, action_id)
    aname = action.name
    db.session.delete(action)

    message = "Action {} deleted".format(aname)
    alert_type = "alert-success"
    try:
        db.session.commit()
    except IntegrityError:
        message = "Action {} is in use in some rules, can not be deleted!".format(aname)
        alert_type = "alert-danger"

    flash(message, alert_type)
    return redirect(url_for("admin.actions"))


@admin.route("/communities")
@auth_required
@admin_required
def communities():
    communities = db.session.query(Community).all()
    return render_template("pages/communities.html", communities=communities)


@admin.route("/community", methods=["GET", "POST"])
@auth_required
@admin_required
def community():
    form = CommunityForm(request.form)

    if request.method == "POST" and form.validate():
        # test if Coomunity name is unique
        exist = get_existing_community(form.name.data)
        if not exist:
            community = Community(
                name=form.name.data,
                comm=form.comm.data,
                larcomm=form.larcomm.data,
                extcomm=form.extcomm.data,
                description=form.description.data,
                as_path=form.as_path.data,
                role_id=form.role_id.data,
            )
            db.session.add(community)
            db.session.commit()
            flash("Community saved", "alert-success")
            return redirect(url_for("admin.communities"))
        else:
            flash(f"Community with name {form.name.data} already exists", "alert-danger")

    community_url = url_for("admin.community")
    return render_template(
        "forms/simple_form.html",
        title="Add new community to Flowspec",
        form=form,
        community_url=community_url,
    )


@admin.route("/community/edit/<int:community_id>", methods=["GET", "POST"])
@auth_required
@admin_required
def edit_community(community_id):
    community = db.session.get(Community, community_id)
    form = CommunityForm(request.form, obj=community)
    if request.method == "POST" and form.validate():
        form.populate_obj(community)
        db.session.commit()
        flash("Community updated")
        return redirect(url_for("admin.communities"))

    community_url = url_for("admin.edit_community", community_id=community.id)
    return render_template(
        "forms/simple_form.html",
        title="Editing {}".format(community.name),
        form=form,
        community_url=community_url,
    )


@admin.route("/community/delete/<int:community_id>", methods=["GET"])
@auth_required
@admin_required
def delete_community(community_id):
    community = db.session.get(Community, community_id)
    aname = community.name
    db.session.delete(community)
    message = "Community {} deleted".format(aname)
    alert_type = "alert-success"
    try:
        db.session.commit()
    except IntegrityError:
        message = "Community {} is in use in some rules, can not be deleted!".format(aname)
        alert_type = "alert-danger"

    flash(message, alert_type)
    return redirect(url_for("admin.communities"))


@admin.route("/set-org-if-zero", methods=["GET"])
@auth_required
@admin_required
def update_set_org():
    # Define the raw SQL update statement
    update_statement = update_statement = text(
        """
        UPDATE organization 
        SET limit_flowspec4 = 0, limit_flowspec6 = 0, limit_rtbh = 0 
        WHERE limit_flowspec4 IS NULL OR limit_flowspec6 IS NULL OR limit_rtbh IS NULL;
    """
    )
    try:
        # Execute the update query
        db.session.execute(update_statement)
        db.session.commit()
    except Exception as e:
        db.session.rollback()
        flash(f"Error updating organizations: {e}", "alert-danger")

    # Get all flowspec records where org_id is NULL (if this is needed)
    models = [Flowspec4, Flowspec6, RTBH, ApiKey, MachineApiKey]
    user_with_multiple_orgs = {}
    for model in models:
        data_records = model.query.filter(model.org_id == 0).all()
        # Loop through each flowspec record and update org_id based on the user's organization
        updated = 0
        for row in data_records:
            orgs = row.user.organization.all()
            if len(orgs) == 1:
                user_org = orgs[0]
                if user_org:
                    row.org_id = user_org.id
                    updated += 1
            else:
                user_with_multiple_orgs[row.user.email] = [org.name for org in orgs]
        # Commit the changes
        try:
            db.session.commit()
        except Exception as e:
            db.session.rollback()
            flash(f"Error updating {model.__name__}: {e}", "alert-danger")

    return render_template("pages/user_list.html", users=user_with_multiple_orgs, updated=updated)



"""Utility functions for models"""

from datetime import datetime
from flowapp import utils
from flowapp.constants import RuleTypes
from flask import current_app

from flowapp.models.rules.whitelist import Whitelist
from .base import db
from .user import User, Role
from .organization import Organization
from .community import Community
from .rules.flowspec import Flowspec4, Flowspec6
from .rules.rtbh import RTBH
from .rules.base import Action


def check_rule_limit(org_id: int, rule_type: RuleTypes) -> bool:
    """
    Check if the organization has reached the rule limit
    :param org_id: integer organization id
    :param rule_type: RuleType rule type
    :return: boolean
    """
    flowspec4_limit = current_app.config.get("FLOWSPEC4_MAX_RULES", 9000)
    flowspec6_limit = current_app.config.get("FLOWSPEC6_MAX_RULES", 9000)
    rtbh_limit = current_app.config.get("RTBH_MAX_RULES", 100000)
    fs4 = db.session.query(Flowspec4).filter_by(rstate_id=1).count()
    fs6 = db.session.query(Flowspec6).filter_by(rstate_id=1).count()
    rtbh = db.session.query(RTBH).filter_by(rstate_id=1).count()

    # check the organization limits
    org = Organization.query.filter_by(id=org_id).first()
    if rule_type == RuleTypes.IPv4 and org.limit_flowspec4 > 0:
        count = db.session.query(Flowspec4).filter_by(org_id=org_id, rstate_id=1).count()
        return count >= org.limit_flowspec4 or fs4 >= flowspec4_limit
    if rule_type == RuleTypes.IPv6 and org.limit_flowspec6 > 0:
        count = db.session.query(Flowspec6).filter_by(org_id=org_id, rstate_id=1).count()
        return count >= org.limit_flowspec6 or fs6 >= flowspec6_limit
    if rule_type == RuleTypes.RTBH and org.limit_rtbh > 0:
        count = db.session.query(RTBH).filter_by(org_id=org_id, rstate_id=1).count()
        return count >= org.limit_rtbh or rtbh >= rtbh_limit

    return False


def check_global_rule_limit(rule_type: RuleTypes) -> bool:
    flowspec4_limit = current_app.config.get("FLOWSPEC4_MAX_RULES", 9000)
    flowspec6_limit = current_app.config.get("FLOWSPEC6_MAX_RULES", 9000)
    rtbh_limit = current_app.config.get("RTBH_MAX_RULES", 100000)
    fs4 = db.session.query(Flowspec4).filter_by(rstate_id=1).count()
    fs6 = db.session.query(Flowspec6).filter_by(rstate_id=1).count()
    rtbh = db.session.query(RTBH).filter_by(rstate_id=1).count()

    # check the global limits if the organization limits are not set

    if rule_type == RuleTypes.IPv4:
        return fs4 >= flowspec4_limit
    if rule_type == RuleTypes.IPv6:
        return fs6 >= flowspec6_limit
    if rule_type == RuleTypes.RTBH:
        return rtbh >= rtbh_limit

    return False


def get_whitelist_model_if_exists(form_data):
    """
    Check if the record in database exist
    ip, mask should match
    expires, rstate_id, user_id, org_id, created, comment can be different
    """
    record = (
        db.session.query(Whitelist)
        .filter(
            Whitelist.ip == form_data["ip"],
            Whitelist.mask == form_data["mask"],
        )
        .first()
    )

    if record:
        return record

    return False


def get_ipv4_model_if_exists(form_data, rstate_id=1):
    """
    Check if the record in database exist
    Source and destination addresses, protocol, flags, action and packet_len should match
    Other fields can be different
    """
    record = (
        db.session.query(Flowspec4)
        .filter(
            Flowspec4.source == form_data["source"],
            Flowspec4.source_mask == form_data["source_mask"],
            Flowspec4.source_port == form_data["source_port"],
            Flowspec4.dest == form_data["dest"],
            Flowspec4.dest_mask == form_data["dest_mask"],
            Flowspec4.dest_port == form_data["dest_port"],
            Flowspec4.protocol == form_data["protocol"],
            Flowspec4.flags == ";".join(form_data["flags"]),
            Flowspec4.packet_len == form_data["packet_len"],
            Flowspec4.action_id == form_data["action"],
            Flowspec4.rstate_id == rstate_id,
        )
        .first()
    )

    if record:
        return record

    return False


def get_ipv6_model_if_exists(form_data, rstate_id=1):
    """
    Check if the record in database exist
    """
    record = (
        db.session.query(Flowspec6)
        .filter(
            Flowspec6.source == form_data["source"],
            Flowspec6.source_mask == form_data["source_mask"],
            Flowspec6.source_port == form_data["source_port"],
            Flowspec6.dest == form_data["dest"],
            Flowspec6.dest_mask == form_data["dest_mask"],
            Flowspec6.dest_port == form_data["dest_port"],
            Flowspec6.next_header == form_data["next_header"],
            Flowspec6.flags == ";".join(form_data["flags"]),
            Flowspec6.packet_len == form_data["packet_len"],
            Flowspec6.action_id == form_data["action"],
            Flowspec6.rstate_id == rstate_id,
        )
        .first()
    )

    if record:
        return record

    return False


def get_rtbh_model_if_exists(form_data):
    """
    Check RTBH rule exist in database
    IPv4, IPv6 and community_id should match
    Rule can be in any state and have different expires, user_id, org_id, created, comment
    """

    record = (
        db.session.query(RTBH)
        .filter(
            RTBH.ipv4 == form_data["ipv4"],
            RTBH.ipv4_mask == form_data["ipv4_mask"],
            RTBH.ipv6 == form_data["ipv6"],
            RTBH.ipv6_mask == form_data["ipv6_mask"],
            RTBH.community_id == form_data["community"],
        )
        .first()
    )

    if record:
        return record

    return False


def insert_users(users):
    """
    inser list of users {name: string, role_id: integer} to db
    """
    for user in users:
        r = Role.query.filter_by(id=user["role_id"]).first()
        o = Organization.query.filter_by(id=user["org_id"]).first()
        u = User(uuid=user["name"])
        u.role.append(r)
        u.organization.append(o)
        db.session.add(u)

    db.session.commit()


def insert_user(
    uuid: str,
    role_ids: list,
    org_ids: list,
    name: str = None,
    phone: str = None,
    email: str = None,
    comment: str = None,
):
    """
    insert new user with multiple roles and organizations
    :param uuid: string unique user id (eppn or similar)
    :param phone: string phone number
    :param name: string user name
    :param email: string email
    :param comment: string comment / notice
    :param role_ids: list of roles
    :param org_ids: list of orgs
    :return: None
    """
    u = User(uuid=uuid, name=name, phone=phone, comment=comment, email=email)

    for role_id in role_ids:
        r = Role.query.filter_by(id=role_id).first()
        u.role.append(r)

    for org_id in org_ids:
        o = Organization.query.filter_by(id=org_id).first()
        u.organization.append(o)

    db.session.add(u)
    db.session.commit()


def get_user_nets(user_id):
    """
    Return list of network ranges for all user organization
    """
    user = db.session.query(User).filter_by(id=user_id).first()
    orgs = user.organization
    result = []
    for org in orgs:
        result.extend(org.arange.split())

    return result


def get_user_orgs_choices(user_id):
    """
    Return list of orgs as choices for form
    """
    user = db.session.query(User).filter_by(id=user_id).first()
    orgs = user.organization

    return [(g.id, g.name) for g in orgs]


def get_user_actions(user_roles):
    """
    Return list of actions based on current user role
    """
    max_role = max(user_roles)
    if max_role == 3:
        actions = db.session.query(Action).order_by("id").all()
    else:
        actions = db.session.query(Action).filter_by(role_id=max_role).order_by("id").all()
    result = [(g.id, g.name) for g in actions]
    return result


def get_user_communities(user_roles):
    """
    Return list of communities based on current user role
    """
    max_role = max(user_roles)
    if max_role == 3:
        communities = db.session.query(Community).order_by("id")
    else:
        communities = db.session.query(Community).filter_by(role_id=max_role).order_by("id")

    return [(g.id, g.name) for g in communities]


def get_existing_action(name=None, command=None):
    """
    return Action with given name or command if the action exists
    return None if action not exists
    :param name: string action name
    :param command: string action command
    :return: action id
    """
    action = Action.query.filter((Action.name == name) | (Action.command == command)).first()
    return action.id if hasattr(action, "id") else None


def get_existing_community(name=None):
    """
    return Community with given name or command if the action exists
    return None if action not exists
    :param name: string action name
    :param command: string action command
    :return: action id
    """
    community = Community.query.filter(Community.name == name).first()
    return community.id if hasattr(community, "id") else None


def _get_flowspec4_rules(rule_state, sort="expires", order="desc", page=1, per_page=50, paginate=False):
    """Get Flowspec4 rules with optional pagination"""

    today = datetime.now()
    comp_func = utils.get_comp_func(rule_state)

    sorter = getattr(Flowspec4, sort, Flowspec4.id)
    sorting = getattr(sorter, order)

    query = db.session.query(Flowspec4)

    if comp_func:
        query = query.filter(comp_func(Flowspec4.expires, today))

    query = query.order_by(sorting())

    if paginate:
        pagination = query.paginate(page=page, per_page=per_page, error_out=False, max_per_page=500)
        return pagination.items, pagination
    else:
        return query.all()


def _get_flowspec6_rules(rule_state, sort="expires", order="desc", page=1, per_page=50, paginate=False):
    """Get Flowspec6 rules with optional pagination"""

    today = datetime.now()
    comp_func = utils.get_comp_func(rule_state)

    sorter = getattr(Flowspec6, sort, Flowspec6.id)
    sorting = getattr(sorter, order)

    query = db.session.query(Flowspec6)

    if comp_func:
        query = query.filter(comp_func(Flowspec6.expires, today))

    query = query.order_by(sorting())

    if paginate:
        pagination = query.paginate(page=page, per_page=per_page, error_out=False, max_per_page=500)
        return pagination.items, pagination
    else:
        return query.all()


def _get_rtbh_rules(rule_state, sort="expires", order="desc", page=1, per_page=50, paginate=False):
    """Get RTBH rules with optional pagination"""

    today = datetime.now()
    comp_func = utils.get_comp_func(rule_state)

    sorter = getattr(RTBH, sort, RTBH.id)
    sorting = getattr(sorter, order)

    query = db.session.query(RTBH)

    if comp_func:
        query = query.filter(comp_func(RTBH.expires, today))

    query = query.order_by(sorting())

    if paginate:
        pagination = query.paginate(page=page, per_page=per_page, error_out=False, max_per_page=500)
        return pagination.items, pagination
    else:
        return query.all()


def _get_whitelist_rules(rule_state, sort="expires", order="desc", page=1, per_page=50, paginate=False):
    """Get Whitelist rules with optional pagination"""

    today = datetime.now()
    comp_func = utils.get_comp_func(rule_state)

    sorter = getattr(Whitelist, sort, Whitelist.id)
    sorting = getattr(sorter, order)

    query = db.session.query(Whitelist)

    if comp_func:
        query = query.filter(comp_func(Whitelist.expires, today))

    query = query.order_by(sorting())

    if paginate:
        pagination = query.paginate(page=page, per_page=per_page, error_out=False, max_per_page=500)
        return pagination.items, pagination
    else:
        return query.all()


# Facade function - keeps backward compatibility and config-based routing
def get_ip_rules(rule_type, rule_state, sort="expires", order="desc", page=1, per_page=50, paginate=False):
    """
    Returns list of rules sorted by sort column ordered asc or desc, with optional pagination.
    This is a facade function that delegates to type-specific handlers.

    Args:
        rule_type: Type of rule ('ipv4', 'ipv6', 'rtbh', 'whitelist')
        rule_state: State filter ('active', 'expired', 'all')
        sort: Column to sort by (default: 'expires')
        order: Sort order 'asc' or 'desc' (default: 'desc')
        page: Page number (1-indexed, default: 1)
        per_page: Number of items per page (default: 50)
        paginate: If True, return (items, pagination) tuple; if False, return all items

    Returns:
        If paginate=True: tuple of (list of rules, pagination object)
        If paginate=False: list of all rules
    """
    # Dispatch to appropriate handler
    handlers = {
        "ipv4": _get_flowspec4_rules,
        "ipv6": _get_flowspec6_rules,
        "rtbh": _get_rtbh_rules,
        "whitelist": _get_whitelist_rules,
    }

    handler = handlers.get(rule_type)

    if handler:
        return handler(rule_state, sort, order, page, per_page, paginate)
    else:
        # Unknown rule type
        if paginate:
            return [], None
        else:
            return []


def get_user_rules_ids(user_id, rule_type):
    """
    Returns list of rule ids belonging to user
    :param user_id: user id
    :param rule_type: ipv4, ipv6 or rtbh
    :return: list
    """

    if rule_type == "ipv4":
        rules4 = db.session.query(Flowspec4.id).filter_by(user_id=user_id).all()
        return [int(x[0]) for x in rules4]

    if rule_type == "ipv6":
        rules6 = db.session.query(Flowspec6.id).order_by(Flowspec6.expires.desc()).all()
        return [int(x[0]) for x in rules6]

    if rule_type == "rtbh":
        rules_rtbh = db.session.query(RTBH.id).filter_by(user_id=user_id).all()
        return [int(x[0]) for x in rules_rtbh]



import pytest
from flowapp.flowspec import translate_sequence, filter_rules_action, check_limit


def test_translate_number():
    """
    tests for x (integer)  to =x
    """
    assert "[=10]" == translate_sequence("10")


def test_raises():
    """
    tests for translator
    """
    with pytest.raises(ValueError):
        translate_sequence("ahoj")


def test_raises_bad_number():
    """
    tests for translator
    """
    with pytest.raises(ValueError):
        translate_sequence("75555")


def test_translate_range():
    """
    tests for x-y  to >=x&<=y
    """
    assert "[>=10&<=20]" == translate_sequence("10-20")


def test_exact_rule():
    """
    test for >=x&<=y to >=x&<=y
    """
    assert "[>=10&<=20]" == translate_sequence(">=10&<=20")


def test_greater_than():
    """
    test for >x to >=x&<=65535
    """
    assert "[>=10&<=65535]" == translate_sequence(">10")


def test_greater_equal_than():
    """
    test for >=x to >=x&<=65535
    """
    assert "[>=10&<=65535]" == translate_sequence(">=10")


def test_lower_than():
    """
    test for <x to >=0&<=0
    """
    assert "[>=0&<=10]" == translate_sequence("<10")


def test_lower_equal_than():
    """
    test for <x to >=0&<=0
    """
    assert "[>=0&<=10]" == translate_sequence("<=10")


# new tests


def test_multiple_sequences():
    """Test multiple sequences separated by semicolons"""
    assert "[=10 >=20&<=30]" == translate_sequence("10;20-30")
    assert "[=10 >=0&<=20 >=30&<=65535]" == translate_sequence("10;<=20;>30")


def test_empty_sequence():
    """Test empty sequence and sequences with empty parts"""
    assert "[]" == translate_sequence("")
    assert "[=10]" == translate_sequence("10;;")
    assert "[=10 =20]" == translate_sequence("10;;20")


def test_range_edge_cases():
    """Test edge cases for ranges"""
    # Same numbers in range
    assert "[>=10&<=10]" == translate_sequence("10-10")

    # Invalid range (start > end)
    with pytest.raises(ValueError, match="Invalid range: start value cannot be greater than end value"):
        translate_sequence("20-10")

    # Invalid range in exact rule
    with pytest.raises(ValueError, match="Invalid range: start value cannot be greater than end value"):
        translate_sequence(">=20&<=10")


def test_check_limit_validation():
    """Test the check_limit function"""
    # Test valid cases
    assert check_limit(10, 100) == 10
    assert check_limit(0, 100, 0) == 0
    assert check_limit(100, 100, 0) == 100

    # Test invalid cases
    with pytest.raises(ValueError, match="Invalid value number: .* is too big"):
        check_limit(101, 100)

    with pytest.raises(ValueError, match="Invalid value number: .* is too small"):
        check_limit(-1, 100, 0)


def test_invalid_inputs():
    """Test various invalid inputs"""
    invalid_inputs = [
        "abc",  # Non-numeric
        "10.5",  # Decimal
        "10,20",  # Wrong separator
        "10&20",  # Wrong format
        ">",  # Incomplete
        ">=",  # Incomplete
        "<",  # Incomplete
        "<=",  # Incomplete
        ">>10",  # Invalid operator
        "<<10",  # Invalid operator
        ">=10&",  # Incomplete range
        "&<=10",  # Incomplete range
    ]

    for invalid_input in invalid_inputs:
        with pytest.raises(ValueError):
            translate_sequence(invalid_input)


class MockRule:
    def __init__(self, action_id):
        self.action_id = action_id


def test_filter_rules_action():
    """Test the filter_rules_action function"""
    # Create test rules
    rules = [MockRule(1), MockRule(2), MockRule(3), MockRule(4)]

    # Test with empty allowed actions
    editable, viewonly = filter_rules_action([], rules)
    assert len(editable) == 0
    assert len(viewonly) == 4

    # Test with some allowed actions
    editable, viewonly = filter_rules_action([1, 3], rules)
    assert len(editable) == 2
    assert len(viewonly) == 2
    assert all(rule.action_id in [1, 3] for rule in editable)
    assert all(rule.action_id in [2, 4] for rule in viewonly)

    # Test with all actions allowed
    editable, viewonly = filter_rules_action([1, 2, 3, 4], rules)
    assert len(editable) == 4
    assert len(viewonly) == 0

    # Test with empty rules list
    editable, viewonly = filter_rules_action([1, 2], [])
    assert len(editable) == 0
    assert len(viewonly) == 0



"""
User-related forms for the flowapp application.
"""

import csv
from io import StringIO

from flask_wtf import FlaskForm
from wtforms import StringField, TextAreaField, SelectMultipleField
from wtforms.validators import ValidationError, DataRequired, Email, InputRequired, Optional


class UserForm(FlaskForm):
    """
    User Form object
    used in Admin
    """

    uuid = StringField(
        "Unique User ID",
        validators=[
            InputRequired("Please provide UUID"),
            Email("Please provide valid email"),
        ],
    )

    email = StringField("Email", validators=[Optional(), Email("Please provide valid email")])

    comment = StringField("Notice", validators=[Optional()])

    name = StringField("Name", validators=[Optional()])

    phone = StringField("Contact phone", validators=[Optional()])

    role_ids = SelectMultipleField("Role", coerce=int, validators=[DataRequired("Select at last one role")])

    org_ids = SelectMultipleField(
        "Organization",
        coerce=int,
        validators=[DataRequired("We prefer one Organization per user, but it's possible select more")],
    )


class BulkUserForm(FlaskForm):
    """
    Bulk User Form object
    used in Admin
    """

    users = TextAreaField("Users in CSV - see example below", validators=[DataRequired()])

    def __init__(self, *args, **kwargs):
        super(BulkUserForm, self).__init__(*args, **kwargs)
        self.roles = None
        self.organizations = None
        self.uuids = None

    # Custom validator for CSV data
    def validate_users(self, field):
        csv_data = field.data

        # Parse CSV data
        csv_reader = csv.DictReader(StringIO(csv_data), delimiter=",")

        # List to keep track of failed validation rows
        errors = 0
        for row_num, row in enumerate(csv_reader, start=1):
            try:
                # check if the user not already exists
                if row["uuid-eppn"] in self.uuids:
                    field.errors.append(f"Row {row_num}: User with UUID {row['uuid-eppn']} already exists.")
                    errors += 1

                # Check if role exists in the database
                role_id = int(row["role"])  # Convert role field to integer
                if role_id not in self.roles:
                    field.errors.append(f"Row {row_num}: Role ID {role_id} does not exist.")
                    errors += 1

                # Check if organization exists in the database
                org_id = int(row["organizace"])  # Convert organization field to integer
                if org_id not in self.organizations:
                    field.errors.append(f"Row {row_num}: Organization ID {org_id} does not exist.")
                    errors += 1

            except (KeyError, ValueError) as e:
                field.errors.append(f"Row {row_num}: Invalid data / key - {str(e)}. Check CSV head row.")

        if errors > 0:
            # Raise validation error if any invalid rows found
            raise ValidationError("Invalid CSV Data - check the errors above.")



import jwt
from flask import (
    Blueprint,
    render_template,
    redirect,
    flash,
    request,
    url_for,
    session,
    make_response,
    current_app,
)
import secrets

from ..forms import ApiKeyForm
from ..models import ApiKey
from ..auth import auth_required

from flowapp import db

COOKIE_KEY = "keylist"

api_keys = Blueprint("api_keys", __name__, template_folder="templates")


@api_keys.route("/", methods=["GET"])
@auth_required
def all():
    """
    Show user api keys
    :return: page with keys
    """
    jwt_key = current_app.config.get("JWT_SECRET")
    keys = db.session.query(ApiKey).filter_by(user_id=session["user_id"]).all()
    payload = {"keys": [key.id for key in keys]}
    encoded = jwt.encode(payload, jwt_key, algorithm="HS256")

    resp = make_response(render_template("pages/api_key.html", keys=keys))

    if current_app.config.get("DEVEL"):
        resp.set_cookie(COOKIE_KEY, encoded, httponly=True, samesite="Lax")
    else:
        resp.set_cookie(COOKIE_KEY, encoded, secure=True, httponly=True, samesite="Lax")

    return resp


@api_keys.route("/add", methods=["GET", "POST"])
@auth_required
def add():
    """
    Add new ApiKey
    :return: form or redirect to list of keys
    """
    generated = secrets.token_hex(24)
    form = ApiKeyForm(request.form, key=generated)

    if request.method == "POST" and form.validate():
        model = ApiKey(
            machine=form.machine.data,
            key=form.key.data,
            expires=form.expires.data,
            readonly=form.readonly.data,
            comment=form.comment.data,
            user_id=session["user_id"],
            org_id=session["user_org_id"],
        )

        db.session.add(model)
        db.session.commit()
        flash("NewKey saved", "alert-success")

        return redirect(url_for("api_keys.all"))
    else:
        for field, errors in form.errors.items():
            for error in errors:
                current_app.logger.debug("Error in the %s field - %s" % (getattr(form, field).label.text, error))

    return render_template("forms/api_key.html", form=form, generated_key=generated)


@api_keys.route("/delete/<int:key_id>", methods=["GET"])
@auth_required
def delete(key_id):
    """
    Delete api_key and machine
    :param key_id: integer
    """
    key_list = request.cookies.get(COOKIE_KEY)
    key_list = jwt.decode(key_list, current_app.config.get("JWT_SECRET"), algorithms=["HS256"])

    model = db.session.get(ApiKey, key_id)
    if model.id not in key_list["keys"]:
        flash("You can't delete this key!", "alert-danger")
    elif model.user_id == session["user_id"] or 3 in session["user_role_ids"]:
        # delete from db
        db.session.delete(model)
        db.session.commit()
        flash("Key deleted", "alert-success")
    else:
        flash("You can't delete this key!", "alert-danger")

    return redirect(url_for("api_keys.all"))



"""
This module contains constant values used in application
"""

from enum import Enum
from operator import ge, lt

DEFAULT_SORT = "expires"
DEFAULT_ORDER = "desc"

# Maximum allowed comma separated values for port string or packet lenght
MAX_COMMA_VALUES = 5

SORT_ARG = "sort"
ORDER_ARG = "order"
RULE_ARG = "rule_state"
TYPE_ARG = "rule_type"
SEARCH_ARG = "squery"
ORDSRC_ARG = "ordsrc"
TIME_FORMAT_ARG = "time_format"
TIME_YEAR = "yearfirst"
TIME_US = "us"
TIME_STMP = "timestamp"

RULES_KEY = "rules"

RULE_TYPES_DICT = {"ipv4": 4, "ipv6": 6, "rtbh": 1}
RULE_NAMES_DICT = {4: "ipv4", 6: "ipv6", 1: "rtbh"}
DEFAULT_COUNT_MATCH = {"ipv4": 0, "ipv6": 0, "rtbh": 0}

ANNOUNCE = 1
WITHDRAW = 2
IPV4_DEFMASK = 32
IPV6_DEFMASK = 128
MAX_PORT = 65535
MAX_PACKET = 9216

IPV6_NEXT_HEADER = {"tcp": "tcp", "udp": "udp", "icmp": "58", "all": ""}

IPV4_PROTOCOL = {"tcp": "tcp", "udp": "udp", "icmp": "icmp", "all": ""}

IPV4_FRAGMENT = {
    "dont": "dont-fragment",
    "first": "first-fragment",
    "is": "is-fragment",
    "last": "last-fragment",
}

COMP_FUNCS = {"active": ge, "expired": lt, "all": None}

TCP_FLAGS = [
    ("SYN", "SYN"),
    ("ACK", "ACK"),
    ("FIN", "FIN"),
    ("RST", "RST"),
    ("PUSH", "PSH"),
    ("URGENT", "URG"),
]

FORM_TIME_PATTERN = "%Y-%m-%dT%H:%M"


class RuleTypes(Enum):
    RTBH = 1
    IPv4 = 4
    IPv6 = 6


class RuleOrigin(Enum):
    USER = 1
    WHITELIST = 2



def test_dashboard_not_auth(client):

    response = client.get("/dashboard/ipv4/active/?sort=expires&order=desc")

    # Expecting a 302 redirect to login
    assert response.status_code == 302


def test_dashboard(auth_client):

    response = auth_client.get("/dashboard/ipv4/active/?sort=expires&order=desc")

    # Check that the request is successful and renders the correct template
    assert response.status_code == 200  # Expecting a 200 OK if the user is authenticated



from sqlalchemy import event
from .base import db


class Organization(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(150), unique=True)
    arange = db.Column(db.Text)
    limit_flowspec4 = db.Column(db.Integer, default=0)
    limit_flowspec6 = db.Column(db.Integer, default=0)
    limit_rtbh = db.Column(db.Integer, default=0)

    def __init__(self, name, arange, limit_flowspec4=0, limit_flowspec6=0, limit_rtbh=0):
        self.name = name
        self.arange = arange
        self.limit_flowspec4 = limit_flowspec4
        self.limit_flowspec6 = limit_flowspec6
        self.limit_rtbh = limit_rtbh

    def __repr__(self):
        return self.name

    def get_users(self):
        """
        Returns all users associated with this organization.
        """
        # self.user is the backref from the user_organization relationship
        return self.user


# Event listeners for Organization
@event.listens_for(Organization.__table__, "after_create")
def insert_initial_organizations(table, conn, *args, **kwargs):
    conn.execute(table.insert().values(name="TU Liberec", arange="147.230.0.0/16\n2001:718:1c01::/48"))
    conn.execute(table.insert().values(name="Cesnet", arange="147.230.0.0/16\n2001:718:1c01::/48"))



from sqlalchemy import event
from .base import db


class Community(db.Model):
    """Community for RTBH rule"""

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(120), unique=True)
    comm = db.Column(db.String(2047))
    larcomm = db.Column(db.String(2047))
    extcomm = db.Column(db.String(2047))
    description = db.Column(db.String(255))
    as_path = db.Column(db.Boolean, default=False)
    role_id = db.Column(db.Integer, db.ForeignKey("role.id"), nullable=False)
    role = db.relationship("Role", backref="community")

    def __init__(self, name, comm, larcomm, extcomm, description, as_path, role_id):
        self.name = name
        self.comm = comm
        self.larcomm = larcomm
        self.extcomm = extcomm
        self.description = description
        self.as_path = as_path
        self.role_id = role_id

    @classmethod
    def get_whitelistable_communities(cls, id_list):
        return cls.query.filter(cls.id.in_(id_list)).all()

    def __repr__(self):
        return f"<Community {self.name}>"

    def __str__(self):
        return f"{self.name}"


class ASPath(db.Model):
    """AS Path for RTBH rules"""

    id = db.Column(db.Integer, primary_key=True)
    prefix = db.Column(db.String(120), unique=True)
    as_path = db.Column(db.String(250))

    # Methods and initializer


@event.listens_for(Community.__table__, "after_create")
def insert_initial_communities(table, conn, *args, **kwargs):
    conn.execute(
        table.insert().values(
            name="65535:65283",
            comm="65535:65283",
            larcomm="",
            extcomm="",
            description="local-as",
            role_id=2,
        )
    )
    conn.execute(
        table.insert().values(
            name="64496:64511",
            comm="64496:64511",
            larcomm="",
            extcomm="",
            description="",
            role_id=2,
        )
    )
    conn.execute(
        table.insert().values(
            name="64497:64510",
            comm="64497:64510",
            larcomm="",
            extcomm="",
            description="",
            role_id=2,
        )
    )



"""
Module for message announcing and logging
"""

from dataclasses import dataclass, asdict
from datetime import datetime
from typing import Dict, Union, Callable, Any

import requests
import pika
import json
from flask import current_app

from flowapp import db, messages
from flowapp.constants import RuleTypes
from flowapp.models import Log
from flowapp.models import Flowspec4, Flowspec6, RTBH


ROUTE_MODELS: Dict[int, Callable[[Any], str]] = {
    1: messages.create_rtbh,
    4: messages.create_ipv4,
    6: messages.create_ipv6,
}


class RouteSources:
    UI: str = "UI"
    API: str = "API"


@dataclass
class Route:
    author: str
    source: str  # Using str instead of RouteSources for flexibility
    command: str

    def __dict__(self) -> Dict[str, str]:
        return asdict(self)


def announce_route(route: Route) -> None:
    """
    Dispatch route as dict to ExaBGP API
    API must be set in app config.py
    defaults to HTTP API
    """
    current_app.logger.debug(asdict(route))
    if current_app.config.get("EXA_API") == "RABBIT":
        announce_to_rabbitmq(asdict(route))
    else:
        announce_to_http(asdict(route))


def announce_to_http(route: Dict[str, str]) -> None:
    """
    Announce route to ExaBGP HTTP API process
    """
    if not current_app.config["TESTING"]:
        try:
            resp = requests.post(current_app.config["EXA_API_URL"], data={"command": json.dumps(route)})
            resp.raise_for_status()
        except requests.exceptions.HTTPError as err:
            current_app.logger.error("ExaAPI HTTP Error: ", err)
        except requests.exceptions.RequestException as ce:
            current_app.logger.error("Connection to ExaAPI failed: ", ce)
    else:
        current_app.logger.debug(f"Testing: {route}")


def announce_to_rabbitmq(route: Dict[str, str]) -> None:
    """
    Announce rout to ExaBGP RabbitMQ API process
    """
    if not current_app.config["TESTING"]:
        user = current_app.config["EXA_API_RABBIT_USER"]
        passwd = current_app.config["EXA_API_RABBIT_PASS"]
        queue = current_app.config["EXA_API_RABBIT_QUEUE"]
        credentials = pika.PlainCredentials(user, passwd)
        parameters = pika.ConnectionParameters(
            current_app.config["EXA_API_RABBIT_HOST"],
            current_app.config["EXA_API_RABBIT_PORT"],
            current_app.config["EXA_API_RABBIT_VHOST"],
            credentials,
        )

        with pika.BlockingConnection(parameters) as connection:
            channel = connection.channel()
            channel.queue_declare(queue=queue)
            channel.basic_publish(exchange="", routing_key=queue, body=json.dumps(route))
    else:
        current_app.logger.debug(f"Testing: {route}")


def log_route(user_id: int, route_model: Union[RTBH, Flowspec4, Flowspec6], rule_type: RuleTypes, author: str) -> None:
    """
    Convert route to EXAFS message and log it to database
    :param user_id : int curent user
    :param route_model: model with route object (RTBH, Flowspec4, or Flowspec6)
    :param rule_type: RuleTypes enum
    :param author: str name of the author
    :return: None
    """
    rule_type_value = rule_type.value
    converter = ROUTE_MODELS[rule_type_value]
    task = converter(route_model)
    log = Log(
        time=datetime.now(),
        task=task,
        rule_type=rule_type_value,
        rule_id=route_model.id,
        user_id=user_id,
        author=author,
    )
    db.session.add(log)
    current_app.logger.info(log)
    db.session.commit()


def log_withdraw(user_id: int, task: str, rule_type: RuleTypes, deleted_id: int, author: str) -> None:
    """
    Log the withdraw command to database
    :param user_id: int user ID
    :param task: str command message
    :param rule_type: RuleTypes enum
    :param deleted_id: int ID of deleted rule
    :param author: str name of the author
    :return: None
    """
    log = Log(
        time=datetime.now(),
        task=task,
        rule_type=rule_type.value,
        rule_id=deleted_id,
        user_id=user_id,
        author=author,
    )
    db.session.add(log)
    current_app.logger.info(log)
    db.session.commit()



from datetime import datetime
from .base import db


class ApiKey(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    machine = db.Column(db.String(255))
    key = db.Column(db.String(255))
    readonly = db.Column(db.Boolean, default=False)
    expires = db.Column(db.DateTime, nullable=True)
    comment = db.Column(db.String(255))
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    user = db.relationship("User", back_populates="apikeys")
    org_id = db.Column(db.Integer, db.ForeignKey("organization.id"), nullable=False)
    org = db.relationship("Organization", backref="apikey")

    def is_expired(self):
        if self.expires is None:
            return False  # Non-expiring key
        else:
            return self.expires < datetime.now()


class MachineApiKey(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    machine = db.Column(db.String(255))
    key = db.Column(db.String(255))
    readonly = db.Column(db.Boolean, default=True)
    expires = db.Column(db.DateTime, nullable=True)
    comment = db.Column(db.String(255))
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    user = db.relationship("User", back_populates="machineapikeys")
    org_id = db.Column(db.Integer, db.ForeignKey("organization.id"), nullable=False)
    org = db.relationship("Organization", backref="machineapikey")

    def is_expired(self):
        if self.expires is None:
            return False  # Non-expiring key
        else:
            return self.expires < datetime.now()



"""
API key forms for the flowapp application.
"""

from flask_wtf import FlaskForm
from wtforms import SelectField, StringField, TextAreaField, BooleanField, HiddenField
from wtforms.validators import DataRequired, IPAddress, Optional, Length

from .base import MultiFormatDateTimeLocalField
from ..constants import FORM_TIME_PATTERN


class ApiKeyForm(FlaskForm):
    """
    ApiKey for User
    Each key / machine pair is unique
    """

    machine = StringField(
        "Machine address",
        validators=[DataRequired(), IPAddress(message="provide valid IP address")],
    )

    comment = TextAreaField("Your comment for this key", validators=[Optional(), Length(max=255)])

    expires = MultiFormatDateTimeLocalField(
        "Key expiration. Leave blank for non expring key (not-recomended).",
        format=FORM_TIME_PATTERN,
        validators=[Optional()],
        unlimited=True,
    )

    readonly = BooleanField("Read only key", default=False)

    key = HiddenField("GeneratedKey")


class MachineApiKeyForm(FlaskForm):
    """
    ApiKey for Services / No login users.
    Each key / machine pair is unique
    This form is used by Admin to create api key for services or users with no Shibboleth login
    User must be created first and must have an organization
    """

    machine = StringField(
        "Machine address",
        validators=[DataRequired(), IPAddress(message="provide valid IP address")],
    )

    comment = TextAreaField("Your comment for this key", validators=[Optional(), Length(max=255)])

    expires = MultiFormatDateTimeLocalField(
        "Key expiration. Leave blank for non expring key (not-recomended).",
        format=FORM_TIME_PATTERN,
        validators=[Optional()],
        unlimited=True,
    )

    readonly = BooleanField("Read only key", default=True)

    user = SelectField(
        "User",
        coerce=int,
        validators=[DataRequired("Select user")],
    )

    key = HiddenField("GeneratedKey")



import json
from datetime import datetime
from flowapp import utils
from ..base import db


class RTBH(db.Model):
    __tablename__ = "RTBH"

    id = db.Column(db.Integer, primary_key=True)
    ipv4 = db.Column(db.String(255))
    ipv4_mask = db.Column(db.Integer)
    ipv6 = db.Column(db.String(255))
    ipv6_mask = db.Column(db.Integer)
    community_id = db.Column(db.Integer, db.ForeignKey("community.id"), nullable=False)
    community = db.relationship("Community", backref="rtbh")
    comment = db.Column(db.Text)
    expires = db.Column(db.DateTime)
    created = db.Column(db.DateTime)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    user = db.relationship("User", backref="rtbh")
    org_id = db.Column(db.Integer, db.ForeignKey("organization.id"), nullable=False)
    org = db.relationship("Organization", backref="rtbh")
    rstate_id = db.Column(db.Integer, db.ForeignKey("rstate.id"), nullable=False)
    rstate = db.relationship("Rstate", backref="RTBH")

    def __init__(
        self,
        ipv4,
        ipv4_mask,
        ipv6,
        ipv6_mask,
        community_id,
        expires,
        user_id,
        org_id,
        comment=None,
        created=None,
        rstate_id=1,
    ):
        self.ipv4 = ipv4
        self.ipv4_mask = ipv4_mask
        self.ipv6 = ipv6
        self.ipv6_mask = ipv6_mask
        self.community_id = community_id
        self.expires = expires
        self.user_id = user_id
        self.org_id = org_id
        self.comment = comment
        if created is None:
            created = datetime.now()
        self.created = created
        self.rstate_id = rstate_id

    def __eq__(self, other):
        """
        Two models are equal if all the network parameters equals.
        User_id and time fields can differ.
        :param other: other RTBH instance
        :return: boolean
        """
        return (
            self.ipv4 == other.ipv4
            and self.ipv4_mask == other.ipv4_mask
            and self.ipv6 == other.ipv6
            and self.ipv6_mask == other.ipv6_mask
            and self.community_id == other.community_id
            and self.rstate_id == other.rstate_id
        )

    def __ne__(self, other):
        """
        Two models are not equal if all the network parameters are not equal.
        User_id and time fields can differ.
        :param other: other RTBH instance
        :return: boolean
        """
        compars = (
            self.ipv4 == other.ipv4
            and self.ipv4_mask == other.ipv4_mask
            and self.ipv6 == other.ipv6
            and self.ipv6_mask == other.ipv6_mask
            and self.community_id == other.community_id
            and self.rstate_id == other.rstate_id
        )

        return not compars

    def update_time(self, form):
        self.expires = utils.webpicker_to_datetime(form.expire_date.data)
        db.session.commit()

    def to_dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict used in API
        :param prefered_format: string with prefered time format
        :return: dictionary
        """
        if prefered_format == "timestamp":
            expires = int(datetime.timestamp(self.expires))
            created = int(datetime.timestamp(self.created))
        else:
            expires = utils.datetime_to_webpicker(self.expires, prefered_format)
            created = utils.datetime_to_webpicker(self.created, prefered_format)

        return {
            "id": self.id,
            "ipv4": self.ipv4,
            "ipv4_mask": self.ipv4_mask,
            "ipv6": self.ipv6,
            "ipv6_mask": self.ipv6_mask,
            "community": self.community.name,
            "comment": self.comment,
            "expires": expires,
            "created": created,
            "user": self.user.uuid,
            "rstate": self.rstate.description,
        }

    def dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :returns: dictionary
        """
        return self.to_dict(prefered_format)

    def json(self, prefered_format="yearfirst"):
        """
        Serialize to json
        :param prefered_format: string with prefered time format
        :returns: json
        """
        return json.dumps(self.to_dict())

    def __repr__(self):
        if not self.ipv6 and not self.ipv6_mask:
            return f"<RTBH {self.ipv4}/{self.ipv4_mask}>"
        if not self.ipv4 and not self.ipv4_mask:
            return f"<RTBH {self.ipv6}/{self.ipv6_mask}>"

        return f"<RTBH {self.ipv4}/{self.ipv4_mask} {self.ipv6}/{self.ipv6_mask}>"

    def __str__(self):
        if not self.ipv6 and not self.ipv6_mask:
            return f"{self.ipv4}/{self.ipv4_mask}"
        if not self.ipv4 and not self.ipv4_mask:
            return f"{self.ipv6}/{self.ipv6_mask}"

        return f"{self.ipv4}/{self.ipv4_mask} {self.ipv6}/{self.ipv6_mask}"

    def get_author(self):
        return f"{self.user.email} / {self.org}"



from .flowspec import Flowspec4, Flowspec6
from .rtbh import RTBH
from .base import Rstate, Action
from .whitelist import Whitelist, RuleWhitelistCache

__all__ = ["Flowspec4", "Flowspec6", "RTBH", "Rstate", "Action", "Whitelist", "RuleWhitelistCache"]



import ipaddress
import pytest

from datetime import datetime, timedelta

from flowapp import utils


@pytest.mark.parametrize(
    "apitime, preformat",
    [
        ("10/15/2015 14:46", "us"),
        ("2015/10/15 14:46", "yearfirst"),
        ("1444913400", "timestamp"),
        (1444913400, "timestamp"),
    ],
)
def test_parse_api_time(apitime, preformat):
    """
    is the time parsed correctly
    """
    result = utils.parse_api_time(apitime)
    assert isinstance(result, tuple)
    assert result[0] == datetime(2015, 10, 15, 14, 50)
    assert result[1] == preformat


@pytest.mark.parametrize(
    "apitime", ["10/152015 14:46", "201/10/15 14:46", "144123254913400", "abcd"]
)
def test_parse_api_time_bad_time(apitime):
    """
    is the time parsed correctly
    """
    assert not utils.parse_api_time(apitime)


def test_get_rule_state_by_time():
    """
    Test if time in the past returns 2
    """
    past = datetime.now() - timedelta(days=1)

    assert utils.get_state_by_time(past) == 2


def test_round_to_ten():
    """
    Test if the time is rounded correctly
    """
    d1 = datetime(2013, 9, 2, 16, 25, 59)
    d2 = datetime(2013, 9, 2, 16, 32, 59)
    dround = datetime(2013, 9, 2, 16, 30, 00)

    assert utils.round_to_ten_minutes(d1) == dround
    assert utils.round_to_ten_minutes(d2) == dround


@pytest.mark.parametrize(
    "address_a, address_b",
    [
        (
            "2001:718:1c01:16:f1c4:c682:817:7e23",
            "2001:0718:1c01:0016:f1c4:c682:0817:7e23",
        ),
        ("2001:718::", "2001:718::0"),
        ("2001:718::0", "2001:0718:0000:0000:0000:0000:0000:0000"),
    ],
)
def test_ipv6_comparsion(address_a, address_b):
    assert ipaddress.ip_address(address_a) == ipaddress.ip_address(address_b)



from datetime import datetime, timedelta
from flowapp.models import (
    User,
    Organization,
    Role,
    ApiKey,
    MachineApiKey,
    Rstate,
    Community,
    Action,
    Flowspec6,
    Whitelist,
)

import flowapp.models as models


def test_insert_ipv4(db):
    """
    test the record can be inserted
    :param db: conftest fixture
    :return:
    """
    model = models.Flowspec4(
        source="192.168.1.1",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        protocol="tcp",
        flags="",
        packet_len="",
        fragment="",
        action_id=1,
        expires=datetime.now(),
        user_id=1,
        org_id=1,
        rstate_id=1,
    )
    db.session.add(model)
    db.session.commit()


def test_get_ipv4_model_if_exists(db):
    """
    test if the function find existing model correctly
    :param db: conftest fixture
    :return:
    """
    model = models.Flowspec4(
        source="192.168.1.1",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        protocol="tcp",
        flags="",
        fragment="",
        packet_len="",
        action_id=1,
        expires=datetime.now(),
        user_id=1,
        org_id=1,
        rstate_id=1,
    )
    db.session.add(model)
    db.session.commit()

    form_data = {
        "source": "192.168.1.1",
        "source_mask": "32",
        "source_port": "80",
        "dest": "",
        "dest_mask": "",
        "dest_port": "",
        "protocol": "tcp",
        "flags": "",
        "packet_len": "",
        "action": 1,
    }

    result = models.get_ipv4_model_if_exists(form_data, 1)
    assert result
    assert result == model


def test_get_ipv6_model_if_exists(db):
    """
    test if the function find existing model correctly
    :param db: conftest fixture
    :return:
    """
    model = models.Flowspec6(
        source="2001:0db8:85a3:0000:0000:8a2e:0370:7334",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        next_header="tcp",
        flags="",
        packet_len="",
        action_id=1,
        expires=datetime.now(),
        user_id=1,
        org_id=1,
        rstate_id=1,
    )
    db.session.add(model)
    db.session.commit()

    form_data = {
        "source": "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
        "source_mask": "32",
        "source_port": "80",
        "dest": "",
        "dest_mask": "",
        "dest_port": "",
        "next_header": "tcp",
        "flags": "",
        "packet_len": "",
        "action": 1,
    }

    result = models.get_ipv6_model_if_exists(form_data, 1)
    assert result
    assert result == model


def test_ipv4_eq(db):
    """
    test that creating with valid data returns 201
    """
    model_A = models.Flowspec4(
        source="192.168.1.1",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        protocol="tcp",
        flags="",
        fragment="",
        packet_len="",
        action_id=1,
        expires="123",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )

    model_B = models.Flowspec4(
        source="192.168.1.1",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        protocol="tcp",
        flags="",
        fragment="",
        packet_len="",
        action_id=1,
        expires="123456",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )

    assert model_A == model_B


def test_ipv4_ne(db):
    """
    test that creating with valid data returns 201
    """
    model_A = models.Flowspec4(
        source="192.168.2.2",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        protocol="tcp",
        flags="",
        fragment="",
        packet_len="",
        action_id=1,
        expires="123",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )

    model_B = models.Flowspec4(
        source="192.168.1.1",
        source_mask="32",
        source_port="80",
        destination="",
        destination_mask="",
        destination_port="",
        protocol="tcp",
        flags="",
        fragment="",
        packet_len="",
        action_id=1,
        expires="123456",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )

    assert model_A != model_B


def test_rtbj_eq(db):
    """
    test that two equal rtbh rules are equal
    """
    model_A = models.RTBH(
        ipv4="192.168.1.1",
        ipv4_mask="32",
        ipv6="",
        ipv6_mask="",
        community_id=1,
        expires="123",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )

    model_B = models.RTBH(
        ipv4="192.168.1.1",
        ipv4_mask="32",
        ipv6="",
        ipv6_mask="",
        community_id=1,
        expires="123456",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )

    assert model_A == model_B


def test_user_creation(db):
    """Test basic user creation and relationships"""
    # Create test role and org first
    role = Role(name="test_role", description="Test Role")
    org = Organization(name="test_org", arange="10.0.0.0/8")
    db.session.add_all([role, org])
    db.session.commit()

    # Create user with relationships
    user = User(
        uuid="test-user-123", name="Test User", phone="1234567890", email="test@example.com", comment="Test comment"
    )
    user.role.append(role)
    user.organization.append(org)
    db.session.add(user)
    db.session.commit()

    # Verify user and relationships
    assert user.uuid == "test-user-123"
    assert user.name == "Test User"
    assert len(user.role.all()) == 1
    assert len(user.organization.all()) == 1
    assert user.role.first().name == "test_role"
    assert user.organization.first().name == "test_org"


def test_api_key_expiration(db):
    """Test ApiKey expiration logic"""
    user = User(uuid="test-user")
    org = Organization(name="test-org", arange="10.0.0.0/8")
    db.session.add_all([user, org])
    db.session.commit()

    # Create non-expiring key
    non_expiring_key = ApiKey(
        machine="test-machine-1",
        key="key1",
        readonly=True,
        expires=None,
        comment="Non-expiring key",
        user_id=user.id,
        org_id=org.id,
    )

    # Create expired key
    expired_key = ApiKey(
        machine="test-machine-2",
        key="key2",
        readonly=True,
        expires=datetime.now() - timedelta(days=1),
        comment="Expired key",
        user_id=user.id,
        org_id=org.id,
    )

    # Create future key
    future_key = ApiKey(
        machine="test-machine-3",
        key="key3",
        readonly=True,
        expires=datetime.now() + timedelta(days=1),
        comment="Future key",
        user_id=user.id,
        org_id=org.id,
    )

    db.session.add_all([non_expiring_key, expired_key, future_key])
    db.session.commit()

    assert not non_expiring_key.is_expired()
    assert expired_key.is_expired()
    assert not future_key.is_expired()


def test_machine_api_key_expiration(db):
    """Test MachineApiKey expiration logic"""
    user = User(uuid="test-user-machine")
    org = Organization(name="test-org-machine", arange="10.0.0.0/8")
    db.session.add_all([user, org])
    db.session.commit()

    # Create non-expiring key
    non_expiring_key = MachineApiKey(
        machine="test-machine-1",
        key="key1",
        readonly=True,
        expires=None,
        comment="Non-expiring key",
        user_id=user.id,
        org_id=org.id,
    )

    # Create expired key
    expired_key = MachineApiKey(
        machine="test-machine-2",
        key="key2",
        readonly=True,
        expires=datetime.now() - timedelta(days=1),
        comment="Expired key",
        user_id=user.id,
        org_id=org.id,
    )

    db.session.add_all([non_expiring_key, expired_key])
    db.session.commit()

    assert not non_expiring_key.is_expired()
    assert expired_key.is_expired()


def test_organization_get_users(db):
    """Test Organization's get_users method"""
    org = Organization(
        name="test-org-get-user", arange="10.0.0.0/8", limit_flowspec4=100, limit_flowspec6=100, limit_rtbh=100
    )
    uuid1 = "test-org-get-user"
    uuid2 = "test-org-get-user2"
    user1 = User(uuid=uuid1)
    user2 = User(uuid=uuid2)

    db.session.add(org)
    db.session.add_all([user1, user2])
    db.session.commit()

    org.user.append(user1)
    org.user.append(user2)
    db.session.commit()

    users = org.get_users()
    assert len(users) == 2
    assert all(isinstance(user, User) for user in users)
    assert {user.uuid for user in users} == {uuid1, uuid2}


def test_flowspec6_equality(db):
    """Test Flowspec6 equality comparison"""
    model_a = Flowspec6(
        source="2001:db8::1",
        source_mask=128,
        source_port="80",
        destination="2001:db8::2",
        destination_mask=128,
        destination_port="443",
        next_header="tcp",
        flags="",
        packet_len="",
        expires=datetime.now(),
        user_id=1,
        org_id=1,
        action_id=1,
    )

    # Same network parameters but different timestamps
    model_b = Flowspec6(
        source="2001:db8::1",
        source_mask=128,
        source_port="80",
        destination="2001:db8::2",
        destination_mask=128,
        destination_port="443",
        next_header="tcp",
        flags="",
        packet_len="",
        expires=datetime.now() + timedelta(days=1),
        user_id=1,
        org_id=1,
        action_id=1,
    )

    # Different network parameters
    model_c = Flowspec6(
        source="2001:db8::3",
        source_mask=128,
        source_port="80",
        destination="2001:db8::4",
        destination_mask=128,
        destination_port="443",
        next_header="tcp",
        flags="",
        packet_len="",
        expires=datetime.now(),
        user_id=1,
        org_id=1,
        action_id=1,
    )

    assert model_a == model_b  # Should be equal despite different timestamps
    assert model_a != model_c  # Should be different due to different network parameters


def test_whitelist_equality(db):
    """Test Whitelist equality comparison"""
    model_a = Whitelist(
        ip="192.168.1.1", mask=32, expires=datetime.now(), user_id=1, org_id=1, comment="Test whitelist"
    )

    # Same IP/mask but different timestamps
    model_b = Whitelist(
        ip="192.168.1.1",
        mask=32,
        expires=datetime.now() + timedelta(days=1),
        user_id=1,
        org_id=1,
        comment="Different comment",
    )

    # Different IP
    model_c = Whitelist(
        ip="192.168.1.2", mask=32, expires=datetime.now(), user_id=1, org_id=1, comment="Test whitelist"
    )

    assert model_a == model_b  # Should be equal despite different timestamps
    assert model_a != model_c  # Should be different due to different IP


def test_whitelist_to_dict(db):
    """Test Whitelist to_dict serialization"""
    whitelist = Whitelist(
        ip="192.168.1.1", mask=32, expires=datetime.now(), user_id=1, org_id=1, comment="Test whitelist"
    )

    # Create required related objects
    user = User(uuid="test-user-whitelist")
    rstate = Rstate(description="active")
    db.session.add_all([user, rstate])
    db.session.commit()

    db.session.add(whitelist)
    db.session.commit()

    whitelist.user = user
    whitelist.rstate_id = rstate.id
    db.session.add(whitelist)
    db.session.commit()

    # Test timestamp format
    dict_timestamp = whitelist.to_dict(prefered_format="timestamp")
    assert isinstance(dict_timestamp["expires"], int)
    assert isinstance(dict_timestamp["created"], int)

    # Test yearfirst format
    dict_yearfirst = whitelist.to_dict(prefered_format="yearfirst")
    assert isinstance(dict_yearfirst["expires"], str)
    assert isinstance(dict_yearfirst["created"], str)

    # Check basic fields
    assert dict_timestamp["ip"] == "192.168.1.1"
    assert dict_timestamp["mask"] == 32
    assert dict_timestamp["comment"] == "Test whitelist"
    assert dict_timestamp["user"] == "test-user-whitelist"



# flowapp/services/rule_service.py
"""
Service module for rule operations.

This module provides business logic functions for creating, updating,
and managing flow rules, separating these concerns from HTTP handling.
"""
from flask import current_app
from typing import Dict, Tuple, List

import sqlalchemy

from flowapp import db
from flowapp.constants import RuleOrigin, RuleTypes
from flowapp.models import Whitelist, RuleWhitelistCache, get_whitelist_model_if_exists
from flowapp.models.rules.flowspec import Flowspec4, Flowspec6
from flowapp.models.rules.rtbh import RTBH
from flowapp.services.base import announce_rtbh_route, withdraw_rtbh_route
from flowapp.services.whitelist_common import add_rtbh_rule_to_cache, create_rtbh_from_whitelist_parts
from flowapp.services.whitelist_common import (
    Relation,
    check_whitelist_against_rules,
    subtract_network,
    whitelist_rtbh_rule,
)
from flowapp.utils import round_to_ten_minutes, quote_to_ent


def create_or_update_whitelist(
    form_data: Dict, user_id: int, org_id: int, user_email: str, org_name: str
) -> Tuple[Whitelist, str]:
    """
    Create a new Whitelist rule or update expiration time of an existing one.

    Args:
        form_data: Validated form data
        user_id: Current user ID
        org_id: Current organization ID
        user_email: User email for logging
        org_name: Organization name for logging

    Returns:
        Tuple containing (whitelist_model, message)
    """
    # Check for existing model
    model = get_whitelist_model_if_exists(form_data)
    flashes = []
    if model:
        model.expires = round_to_ten_minutes(form_data["expires"])
        flashes.append("Existing Whitelist found. Expiration time was updated to new value.")
    else:
        # Create new model
        model = Whitelist(
            ip=form_data["ip"],
            mask=form_data["mask"],
            expires=round_to_ten_minutes(form_data["expires"]),
            user_id=user_id,
            org_id=org_id,
            comment=quote_to_ent(form_data["comment"]),
        )
        db.session.add(model)
        flashes.append("Whitelist saved")

    db.session.commit()

    # check RTBH rules against whitelist
    allowed_communities = current_app.config["ALLOWED_COMMUNITIES"]
    # filter out RTBH rules that are not active or whitelisted and not in allowed communities
    all_rtbh_rules = RTBH.query.filter(RTBH.rstate_id.in_([1, 4]), RTBH.community_id.in_(allowed_communities)).all()
    rtbh_rules_map = map_rtbh_rules_to_strings(all_rtbh_rules)
    result = check_whitelist_against_rules(rtbh_rules_map, str(model))
    current_app.logger.info(f"Found {len(result)} matching RTBH rules for whitelist {model}")
    model = evaluate_whitelist_against_rtbh_check_results(model, flashes, rtbh_rules_map, result)

    return model, flashes


def evaluate_whitelist_against_rtbh_check_results(
    whitelist_model: Whitelist,
    flashes: List[str],
    rtbh_rule_cache: Dict[str, Whitelist],
    results: List[Tuple[str, str, Relation]],
) -> Whitelist:

    for rule_key, whitelist_key, relation in results:
        current_app.logger.info(f"whitelist {whitelist_key} is {relation} to Rule {rule_key}")
        if relation == Relation.EQUAL:
            whitelist_rtbh_rule(rtbh_rule_cache[rule_key], whitelist_model)
            withdraw_rtbh_route(rtbh_rule_cache[rule_key])
            msg = "Existing active rule {rule_key} is equal to whitelist {whitelist_key}. Rule is now whitelisted."
            flashes.append(msg)
            current_app.logger.info(msg)
        elif relation == Relation.SUBNET:
            parts = subtract_network(target=rule_key, whitelist=whitelist_key)
            wl_id = whitelist_model.id
            msg = f"Rule {rule_key} is supernet of whitelist {whitelist_key}. Rule is whitelisted, {len(parts)} subnet rules will be created."
            flashes.append(msg)
            current_app.logger.info(msg)
            for network in parts:
                rule_model = rtbh_rule_cache[rule_key]
                create_rtbh_from_whitelist_parts(rule_model, wl_id, whitelist_key, network)
                msg = f"Created RTBH rule from {rule_model.id} {network} parted by whitelist {whitelist_key}."
                flashes.append(msg)
                current_app.logger.info(msg)
            rule_model.rstate_id = 4
            add_rtbh_rule_to_cache(rule_model, wl_id, RuleOrigin.USER)
            db.session.commit()
        elif relation == Relation.SUPERNET:

            whitelist_rtbh_rule(rtbh_rule_cache[rule_key], whitelist_model)
            withdraw_rtbh_route(rtbh_rule_cache[rule_key])
            msg = f"Existing active rule {rule_key} is subnet of whitelist {whitelist_key}. Rule is now whitelisted."
            current_app.logger.info(msg)
            flashes.append(msg)

    return whitelist_model


def map_rtbh_rules_to_strings(all_rtbh_rules: List[RTBH]) -> Dict[str, RTBH]:
    return {str(rule): rule for rule in all_rtbh_rules}


def delete_expired_whitelists() -> List[str]:
    """
    Delete all expired whitelist entries from the database.

    Returns:
        List of messages for the user
    """
    expired_whitelists = Whitelist.query.filter(Whitelist.expires < db.func.now()).all()
    flashes = []
    for model in expired_whitelists:
        flashes.extend(delete_whitelist(model.id))
    return flashes


def delete_whitelist(whitelist_id: int) -> List[str]:
    """
    Delete a whitelist entry from the database.

    Args:
        whitelist_id: The ID of the whitelist to delete
    """
    model = db.session.get(Whitelist, whitelist_id)
    flashes = []
    if model:
        cached_rules = RuleWhitelistCache.get_by_whitelist_id(whitelist_id)
        current_app.logger.info(
            f"Deleting whitelist {whitelist_id} {model}. Found {len(cached_rules)} cached rules to process."
        )
        for cached_rule in cached_rules:
            rule_model_type = RuleTypes(cached_rule.rtype)
            cache_entries_count = RuleWhitelistCache.count_by_rule(cached_rule.rid, rule_model_type)
            if rule_model_type == RuleTypes.IPv4:
                rule_model = db.session.get(Flowspec4, cached_rule.rid)
            elif rule_model_type == RuleTypes.IPv6:
                rule_model = db.session.get(Flowspec6, cached_rule.rid)
            elif rule_model_type == RuleTypes.RTBH:
                rule_model = db.session.get(RTBH, cached_rule.rid)
            rorigin_type = RuleOrigin(cached_rule.rorigin)
            current_app.logger.debug(f"Rule {rule_model} has origin {rorigin_type}")
            if rorigin_type == RuleOrigin.WHITELIST:
                msg = f"Deleted {rule_model_type} rule {rule_model} created by whitelist {model}"
                current_app.logger.info(msg)
                flashes.append(msg)
                try:
                    db.session.delete(rule_model)
                except sqlalchemy.orm.exc.UnmappedInstanceError:
                    current_app.logger.warning(
                        f"RuleWhitelistCache Anomaly! Rule {rule_model} does not exist. Type {rule_model_type} RID {cached_rule.rid} ID {cached_rule.id} Skipping."
                    )

            elif rorigin_type == RuleOrigin.USER and cache_entries_count == 1:
                msg = f"Set rule {rule_model} back to state 'Active'"
                current_app.logger.info(msg)
                flashes.append(msg)
                try:
                    rule_model.rstate_id = 1  # Set rule state to "Active" again
                except AttributeError:
                    current_app.logger.warning(
                        f"RuleWhitelistCache Anomaly! Rule {rule_model} does not exist. Type {rule_model_type} RID {cached_rule.rid} ID {cached_rule.id} Skipping."
                    )
                else:
                    author = f"{model.user.email} ({model.user.organization})"
                    announce_rtbh_route(rule_model, author)

        msg = f"Deleted cache entries for whitelist {whitelist_id} {model}"
        current_app.logger.info(msg)
        flashes.append(msg)
        RuleWhitelistCache.clean_by_whitelist_id(whitelist_id)

        db.session.delete(model)
        db.session.commit()
        msg = f"Deleted whitelist {whitelist_id} {model}"
        flashes.append(msg)
        current_app.logger.info(msg)

    return flashes



from datetime import datetime, timedelta
from flask import Blueprint, current_app, flash, redirect, render_template, request, session, url_for

from flowapp.auth import (
    auth_required,
    user_or_admin_required,
)
from flowapp import constants, db
from flowapp.forms import WhitelistForm
from flowapp.models import get_user_nets, Whitelist
from flowapp.services import create_or_update_whitelist, delete_whitelist
from flowapp.utils.base import flash_errors
from flowapp.auth import check_user_can_modify_rule


whitelist = Blueprint("whitelist", __name__, template_folder="templates")


@whitelist.route("/add", methods=["GET", "POST"])
@auth_required
@user_or_admin_required
def add():
    net_ranges = get_user_nets(session["user_id"])
    form = WhitelistForm(request.form)

    form.net_ranges = net_ranges

    if request.method == "POST" and form.validate():
        model, messages = create_or_update_whitelist(
            form.data,
            user_id=session["user_id"],
            org_id=session["user_org_id"],
            user_email=session["user_email"],
            org_name=session["user_org"],
        )
        for message in messages:
            flash(message, "alert-success")

        return redirect(url_for("index"))
    else:
        for field, errors in form.errors.items():
            for error in errors:
                current_app.logger.debug("Error in the %s field - %s" % (getattr(form, field).label.text, error))

    default_expires = datetime.now() + timedelta(hours=1)
    form.expires.data = default_expires

    return render_template("forms/whitelist.html", form=form, action_url=url_for("whitelist.add"))


@whitelist.route("/reactivate/<int:wl_id>", methods=["GET", "POST"])
@auth_required
@user_or_admin_required
def reactivate(wl_id):
    """
    Set new time for whitelist
    :param wl_id: int - id of the whitelist
    """

    model = db.session.get(Whitelist, wl_id)
    form = WhitelistForm(request.form, obj=model)
    form.net_ranges = get_user_nets(session["user_id"])

    # do not need to validate - all is readonly
    if request.method == "POST":
        model = create_or_update_whitelist(
            form.data,
            user_id=session["user_id"],
            org_id=session["user_org_id"],
            user_email=session["user_email"],
            org_name=session["user_org"],
        )
        flash("Whitelist updated", "alert-success")
        return redirect(
            url_for(
                "dashboard.index",
                rtype=session[constants.TYPE_ARG],
                rstate=session[constants.RULE_ARG],
                sort=session[constants.SORT_ARG],
                squery=session[constants.SEARCH_ARG],
                order=session[constants.ORDER_ARG],
            )
        )
    else:
        flash_errors(form)

    form.expires.data = model.expires
    for field in form:
        if field.name not in ["expires", "csrf_token", "comment"]:
            field.render_kw = {"disabled": "disabled"}

    action_url = url_for("whitelist.reactivate", wl_id=wl_id)

    return render_template(
        "forms/whitelist.html",
        form=form,
        action_url=action_url,
        editing=True,
        title="Update",
    )


@whitelist.route("/delete/<int:wl_id>", methods=["GET"])
@auth_required
@user_or_admin_required
def delete(wl_id):
    """
    Delete whitelist
    :param wl_id: integer - id of the whitelist
    """
    # Check if user can modify this whitelist
    if check_user_can_modify_rule(wl_id, "whitelist"):
        messages = delete_whitelist(wl_id)
        flash(f"Whitelist {wl_id} deleted", "alert-success")
        for message in messages:
            flash(message, "alert-info")
    else:
        flash("You can not delete this Whitelist", "alert-warning")

    return redirect(
        url_for(
            "dashboard.index",
            rtype=session[constants.TYPE_ARG],
            rstate=session[constants.RULE_ARG],
            sort=session[constants.SORT_ARG],
            squery=session[constants.SEARCH_ARG],
            order=session[constants.ORDER_ARG],
        )
    )



"""
Forms module for the flowapp application.
This file imports and re-exports all form classes from the module.
"""

# Base field
from .base import MultiFormatDateTimeLocalField

# User forms
from .user import UserForm, BulkUserForm

# API key forms
from .api import ApiKeyForm, MachineApiKeyForm

# Organization forms
from .organization import OrganizationForm

# Action, ASPath, and Community forms
from .choices import ActionForm, ASPathForm, CommunityForm

# Rule forms
from .rules import IPForm, IPv4Form, IPv6Form, RTBHForm, WhitelistForm


__all__ = [
    "MultiFormatDateTimeLocalField",
    "UserForm",
    "BulkUserForm",
    "ApiKeyForm",
    "MachineApiKeyForm",
    "OrganizationForm",
    "ActionForm",
    "ASPathForm",
    "CommunityForm",
    "RTBHForm",
    "IPForm",
    "IPv4Form",
    "IPv6Form",
    "WhitelistForm",
]



from flask import Blueprint, request
from flowapp import csrf
from flowapp.views import api_common

api = Blueprint("api_v3", __name__, template_folder="templates")


@api.route("/auth", methods=["GET"])
def authorize():
    mkey = request.headers.get("x-api-key", None)
    return api_common.authorize(mkey)


@api.route("/rules")
@api_common.token_required
def index(current_user):
    key_map = {
        "ipv4_rules": "flowspec_ipv4_rw",
        "ipv6_rules": "flowspec_ipv6_rw",
        "rtbh_rules": "rtbh_any_rw",
        "ipv4_rules_readonly": "flowspec_ipv4_ro",
        "ipv6_rules_readonly": "flowspec_ipv6_ro",
        "rtbh_rules_readonly": "rtbh_any_ro",
    }
    return api_common.index(current_user, key_map)


@api.route("/actions")
@api_common.token_required
def all_actions(current_user):
    """
    Returns Actions allowed for current user
    :param current_user:
    :return: json response
    """
    return api_common.all_actions(current_user)


@api.route("/communities")
@api_common.token_required
def all_communities(current_user):
    """
    Returns RTHB communites allowed for current user
    :param current_user:
    :return: json response
    """
    return api_common.all_communities(current_user)


@api.route("/rules/ipv4", methods=["POST"])
@api_common.token_required
@api_common.check_readonly
def create_ipv4(current_user):
    """
    Api method for new IPv4 rule
    :param data: parsed json request
    :param current_user: data from jwt token
    :return: json response
    """
    return api_common.create_ipv4(current_user)


@api.route("/rules/ipv6", methods=["POST"])
@csrf.exempt
@api_common.token_required
@api_common.check_readonly
def create_ipv6(current_user):
    """
    Create new IPv6 rule
    :param data: parsed json request
    :param current_user: data from jwt token
    :return:
    """
    return api_common.create_ipv6(current_user)


@api.route("/rules/rtbh", methods=["POST"])
@csrf.exempt
@api_common.token_required
@api_common.check_readonly
def create_rtbh(current_user):
    return api_common.create_rtbh(current_user)


@api.route("/rules/ipv4/<int:rule_id>", methods=["GET"])
@api_common.token_required
def ipv4_rule_get(current_user, rule_id):
    """
    Return IPv4 rule
    :param current_user:
    :param rule_id:
    :return:
    """
    return api_common.ipv4_rule_get(current_user, rule_id)


@api.route("/rules/ipv6/<int:rule_id>", methods=["GET"])
@api_common.token_required
def ipv6_rule_get(current_user, rule_id):
    """
    Return IPv6 rule
    :param current_user:
    :param rule_id:
    :return:
    """
    return api_common.ipv6_rule_get(current_user, rule_id)


@api.route("/rules/rtbh/<int:rule_id>", methods=["GET"])
@api_common.token_required
def rtbh_rule_get(current_user, rule_id):
    """
    Return RTBH rule
    :param current_user:
    :param rule_id:
    :return:
    """
    return api_common.rtbh_rule_get(current_user, rule_id)


@api.route("/rules/ipv4/<int:rule_id>", methods=["DELETE"])
@api_common.token_required
@api_common.check_readonly
def delete_v4_rule(current_user, rule_id):
    """
    Delete rule with given id and type
    :param rule_id: integer - rule id
    """
    return api_common.delete_v4_rule(current_user, rule_id)


@api.route("/rules/ipv6/<int:rule_id>", methods=["DELETE"])
@api_common.token_required
@api_common.check_readonly
def delete_v6_rule(current_user, rule_id):
    """
    Delete rule with given id and type
    :param rule_id: integer - rule id
    """
    return api_common.delete_v6_rule(current_user, rule_id)


@api.route("/rules/rtbh/<int:rule_id>", methods=["DELETE"])
@api_common.token_required
@api_common.check_readonly
def delete_rtbh_rule(current_user, rule_id):
    """
    Delete rule with given id and type
    :param rule_id: integer - rule id
    """
    return api_common.delete_rtbh_rule(current_user, rule_id)


@api.route("/test_token", methods=["GET"])
@api_common.token_required
def token_test_get(current_user):
    """
    Return IPv4 rule
    :param current_user:
    :param rule_id:
    :return:
    """
    return api_common.token_test_get(current_user)



from enum import Enum, auto
from functools import lru_cache
import ipaddress
from typing import List, Tuple, Union
from flowapp import db
from flowapp.constants import RuleOrigin, RuleTypes
from flowapp.models import RTBH, RuleWhitelistCache, Whitelist
from flowapp.output import log_route
from flowapp.services.base import announce_rtbh_route


def add_rtbh_rule_to_cache(model: RTBH, whitelist_id: int, rule_origin: RuleOrigin = RuleOrigin.USER) -> None:
    """
    Add RTBH rule to whitelist cache
    """
    cache = RuleWhitelistCache(rid=model.id, rtype=RuleTypes.RTBH, whitelist_id=whitelist_id, rorigin=rule_origin)
    db.session.add(cache)
    db.session.commit()


def whitelist_rtbh_rule(model: RTBH, whitelist: Whitelist) -> RTBH:
    """
    Whitelist RTBH rule.
    Set rule state to 4 - whitelisted rule, do not announce later
    Add to whitelist cache
    """
    model.rstate_id = 4
    db.session.commit()
    add_rtbh_rule_to_cache(model, whitelist.id, RuleOrigin.USER)
    return model


class Relation(Enum):
    """
    Enum to represent the relation between Whitelist to Rule Relation
    Subnet: Whitelist is a subnet of the rule
    Supernet: Whitelist is a supernet of the rule
    Equal: Whitelist is equal to the rule
    Different: Whitelist is different from the rule
    """

    SUBNET = auto()
    SUPERNET = auto()
    EQUAL = auto()
    DIFFERENT = auto()


def _is_same_ip_version(addr1: str, addr2: str) -> bool:
    """
    Check if two IP addresses/networks are of the same IP version.

    :param addr1: First IP address or network string
    :param addr2: Second IP address or network string
    :return: True if both addresses are of the same IP version (IPv4 or IPv6), False otherwise
    """
    is_ipv4_1 = "." in addr1
    is_ipv4_2 = "." in addr2
    return is_ipv4_1 == is_ipv4_2


@lru_cache(maxsize=1024)
def get_network(address: str) -> Union[ipaddress.IPv4Network, ipaddress.IPv6Network]:
    """
    Create and cache an IP network object.

    :param address: IP address or network in string format
    :return: Cached IP network object
    """
    return ipaddress.ip_network(address, strict=False)


def check_whitelist_to_rule_relation(rule: str, whitelist_entry: str) -> Relation:
    """
    Checks if the whitelist network is a subnet or supernet or exactly the same as the rule network.
    Uses cached network objects for better performance.

    If the rule and whitelist are different IP versions (IPv4 vs IPv6),
    they are treated as different networks.

    :param rule: The IP address or network to check (e.g., "192.168.1.1" or "192.168.1.0/24")
    :param whitelist_entry: The allowed network to compare against (e.g., "192.168.1.0/24")
    :return: Relation between the two networks
    """
    # First check if IP versions are the same
    if not _is_same_ip_version(rule, whitelist_entry):
        return Relation.DIFFERENT

    try:
        rule_net = get_network(rule)
        whitelist_net = get_network(whitelist_entry)

        if whitelist_net == rule_net:
            return Relation.EQUAL
        if whitelist_net.supernet_of(rule_net):
            return Relation.SUPERNET
        if whitelist_net.subnet_of(rule_net):
            return Relation.SUBNET
        else:
            return Relation.DIFFERENT
    except (ValueError, TypeError):
        # Handle any other errors that might occur during comparison
        return Relation.DIFFERENT


def subtract_network(target: str, whitelist: str) -> List[str]:
    """
    Computes the remaining parts of a network after removing the whitelist subnet.
    Uses cached network objects for better performance.

    If the target and whitelist are different IP versions (IPv4 vs IPv6),
    the original target is returned unchanged.

    :param target: The main network (e.g., "192.168.1.0/24")
    :param whitelist: The subnet to remove (e.g., "192.168.1.128/25")
    :return: A list of remaining subnets as strings
    """
    # First check if IP versions are the same
    if not _is_same_ip_version(target, whitelist):
        return [target]

    try:
        target_net = get_network(target)
        whitelist_net = get_network(whitelist)

        # Check if the whitelist is actually a subnet
        if check_whitelist_to_rule_relation(target, whitelist) != Relation.SUBNET:
            return [target]  # Return the full network if whitelist isn't a valid subnet

        remaining = []

        # Compute ranges before and after the whitelist
        if whitelist_net.network_address > target_net.network_address:
            # Before the whitelist
            start = target_net.network_address
            end = whitelist_net.network_address - 1
            remaining.extend(ipaddress.summarize_address_range(start, end))

        if whitelist_net.broadcast_address < target_net.broadcast_address:
            # After the whitelist
            start = whitelist_net.broadcast_address + 1
            end = target_net.broadcast_address
            remaining.extend(ipaddress.summarize_address_range(start, end))

        # Convert to string format
        return [str(net) for net in remaining]
    except (ValueError, TypeError):
        # Return the original target in case of any error
        return [target]


def check_rule_against_whitelists(rule: str, whitelists: List[str]) -> List[Tuple]:
    """
    Helper function to check a single rule against multiple whitelist entries.
    Creates a cached rule network object for better performance.
    Reduces list of whitelists, where the Relation is not DIFFERENT

    :param rule: The IP address or network to check
    :param whitelists: List of whitelist networks to check against
    :return: tuple of rule, whitelist and relation for each whitelists that is not DIFFERENT
    """
    # Pre-cache the rule network since it will be used multiple times
    try:
        get_network(rule)
    except (ValueError, TypeError):
        # Return empty list if rule is not a valid network
        return []

    items = []
    for whitelist in whitelists:
        try:
            rel = check_whitelist_to_rule_relation(rule, whitelist)
            if rel != Relation.DIFFERENT:
                items.append((rule, whitelist, rel))
        except (ValueError, TypeError):
            # Skip this whitelist if there's an error comparing
            continue
    return items


def check_whitelist_against_rules(rules: List[str], whitelist: str) -> List[Tuple]:
    """
    Helper function to check if any whitelist entry is a subnet of the rule.
    Creates a cached rule network object for better performance.
    Reduces list of rules, where the Relation is not DIFFERENT

    :param rules: List of rule networks to check against
    :param whitelist: The whitelist network to check
    :return: tuple of rule, whitelist and relation for each rules that is not DIFFERENT
    """
    # Pre-cache the whitelist network since it will be used multiple times
    try:
        get_network(whitelist)
    except (ValueError, TypeError):
        # Return empty list if whitelist is not a valid network
        return []

    items = []
    for rule in rules:
        try:
            rel = check_whitelist_to_rule_relation(rule, whitelist)
            if rel != Relation.DIFFERENT:
                items.append((rule, whitelist, rel))
        except (ValueError, TypeError):
            # Skip this rule if there's an error comparing
            continue
    return items


def clear_network_cache() -> None:
    """
    Clear the network object cache.
    Useful when processing a large number of networks to prevent memory growth.
    """
    get_network.cache_clear()


def create_rtbh_from_whitelist_parts(
    model: RTBH, wl_id: int, whitelist_key: str, network: str, rule_owner: str = "", user_id: int = 0
) -> RTBH:
    # default values from model
    rule_owner = rule_owner or model.get_author()
    user_id = user_id or model.user_id

    net_ip, net_mask = network.split("/")
    new_model = RTBH(
        ipv4=net_ip,
        ipv4_mask=net_mask,
        ipv6=model.ipv6,
        ipv6_mask=model.ipv6_mask,
        community_id=model.community_id,
        expires=model.expires,
        comment=model.comment,
        user_id=model.user_id,
        org_id=model.org_id,
        rstate_id=1,
    )
    db.session.add(new_model)
    db.session.commit()

    add_rtbh_rule_to_cache(new_model, wl_id, RuleOrigin.WHITELIST)
    announce_rtbh_route(new_model, rule_owner)
    log_route(user_id, model, RuleTypes.RTBH, rule_owner)

    return new_model



"""
Action, ASPath, and Community forms for the flowapp application.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, SelectField, BooleanField
from wtforms.validators import Length, DataRequired


class ActionForm(FlaskForm):
    """
    Action form object
    used in Admin
    """

    name = StringField("Action short name", validators=[Length(max=150)])

    command = StringField("ExaBGP command", validators=[Length(max=150)])

    description = StringField("Action description")

    role_id = SelectField(
        "Minimal required role",
        choices=[("2", "user"), ("3", "admin")],
        validators=[DataRequired()],
    )


class ASPathForm(FlaskForm):
    """
    AS Path form object
    used in Admin
    """

    prefix = StringField("Prefix", validators=[Length(max=120), DataRequired()])

    as_path = StringField("as-path value", validators=[Length(max=250), DataRequired()])


class CommunityForm(FlaskForm):
    """
    Community form object
    used in Admin
    """

    name = StringField("Community short name", validators=[Length(max=120), DataRequired()])

    comm = StringField("Community value", validators=[Length(max=2046)])

    larcomm = StringField("Large community value", validators=[Length(max=2046)])

    extcomm = StringField("Extended community value", validators=[Length(max=2046)])

    description = StringField("Community description", validators=[Length(max=255)])

    role_id = SelectField(
        "Minimal required role",
        choices=[("2", "user"), ("3", "admin")],
        validators=[DataRequired()],
    )

    as_path = BooleanField("add AS-path (checked = true)")

    def validate(self):
        """
        custom validation method
        :return: boolean
        """
        result = True

        if not FlaskForm.validate(self):
            result = False

        if not self.comm.data and not self.extcomm.data and not self.larcomm.data:
            err_message = "At last one of those values could not be empty"
            self.comm.errors.append(err_message)
            self.larcomm.errors.append(err_message)
            self.extcomm.errors.append(err_message)
            result = False

        return result



import json
from datetime import datetime
from flowapp import utils
from ..base import db


class Flowspec4(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    source = db.Column(db.String(255))
    source_mask = db.Column(db.Integer)
    source_port = db.Column(db.String(255))
    dest = db.Column(db.String(255))
    dest_mask = db.Column(db.Integer)
    dest_port = db.Column(db.String(255))
    protocol = db.Column(db.String(255))
    flags = db.Column(db.String(255))
    packet_len = db.Column(db.String(255))
    fragment = db.Column(db.String(255))
    comment = db.Column(db.Text)
    expires = db.Column(db.DateTime)
    created = db.Column(db.DateTime)
    action_id = db.Column(db.Integer, db.ForeignKey("action.id"), nullable=False)
    action = db.relationship("Action", backref="flowspec4")
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    user = db.relationship("User", backref="flowspec4")
    org_id = db.Column(db.Integer, db.ForeignKey("organization.id"), nullable=False)
    org = db.relationship("Organization", backref="flowspec4")
    rstate_id = db.Column(db.Integer, db.ForeignKey("rstate.id"), nullable=False)
    rstate = db.relationship("Rstate", backref="flowspec4")

    def __init__(
        self,
        source,
        source_mask,
        source_port,
        destination,
        destination_mask,
        destination_port,
        protocol,
        flags,
        packet_len,
        fragment,
        expires,
        user_id,
        org_id,
        action_id,
        created=None,
        comment=None,
        rstate_id=1,
    ):
        self.source = source
        self.source_mask = source_mask
        self.dest = destination
        self.dest_mask = destination_mask
        self.source_port = source_port
        self.dest_port = destination_port
        self.protocol = protocol
        self.flags = flags
        self.packet_len = packet_len
        self.fragment = fragment
        self.comment = comment
        self.expires = expires
        self.user_id = user_id
        self.org_id = org_id
        self.action_id = action_id
        if created is None:
            created = datetime.now()
        self.created = created
        self.rstate_id = rstate_id

    def __eq__(self, other):
        """
        Two models are equal if all the network parameters equals.
        User_id and time fields can differ.
        :param other: other Flowspec4 instance
        :return: boolean
        """
        return (
            self.source == other.source
            and self.source_mask == other.source_mask
            and self.dest == other.dest
            and self.dest_mask == other.dest_mask
            and self.source_port == other.source_port
            and self.dest_port == other.dest_port
            and self.protocol == other.protocol
            and self.flags == other.flags
            and self.packet_len == other.packet_len
            and self.fragment == other.fragment
            and self.action_id == other.action_id
            and self.rstate_id == other.rstate_id
        )

    def __ne__(self, other):
        """
        Two models are not equal if all the network parameters are not equal.
        User_id and time fields can differ.
        :param other: other Flowspec4 instance
        :return: boolean
        """
        compars = (
            self.source == other.source
            and self.source_mask == other.source_mask
            and self.dest == other.dest
            and self.dest_mask == other.dest_mask
            and self.source_port == other.source_port
            and self.dest_port == other.dest_port
            and self.protocol == other.protocol
            and self.flags == other.flags
            and self.packet_len == other.packet_len
            and self.fragment == other.fragment
            and self.action_id == other.action_id
            and self.rstate_id == other.rstate_id
        )

        return not compars

    def to_dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :return: dictionary
        """
        if prefered_format == "timestamp":
            expires = int(datetime.timestamp(self.expires))
            created = int(datetime.timestamp(self.created))
        else:
            expires = utils.datetime_to_webpicker(self.expires, prefered_format)
            created = utils.datetime_to_webpicker(self.created, prefered_format)

        return {
            "id": self.id,
            "source": self.source,
            "source_mask": self.source_mask,
            "source_port": self.source_port,
            "dest": self.dest,
            "dest_mask": self.dest_mask,
            "dest_port": self.dest_port,
            "protocol": self.protocol,
            "flags": self.flags,
            "packet_len": self.packet_len,
            "fragment": self.fragment,
            "comment": self.comment,
            "expires": expires,
            "created": created,
            "action": self.action.name,
            "user": self.user.uuid,
            "rstate": self.rstate.description,
        }

    def dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :returns: dictionary
        """
        return self.to_dict(prefered_format)

    def json(self, prefered_format="yearfirst"):
        """
        Serialize to json
        :param prefered_format: string with prefered time format
        :returns: json
        """
        return json.dumps(self.to_dict())


class Flowspec6(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    source = db.Column(db.String(255))
    source_mask = db.Column(db.Integer)
    source_port = db.Column(db.String(255))
    dest = db.Column(db.String(255))
    dest_mask = db.Column(db.Integer)
    dest_port = db.Column(db.String(255))
    next_header = db.Column(db.String(255))
    flags = db.Column(db.String(255))
    packet_len = db.Column(db.String(255))
    comment = db.Column(db.Text)
    expires = db.Column(db.DateTime)
    created = db.Column(db.DateTime)
    action_id = db.Column(db.Integer, db.ForeignKey("action.id"), nullable=False)
    action = db.relationship("Action", backref="flowspec6")
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    user = db.relationship("User", backref="flowspec6")
    org_id = db.Column(db.Integer, db.ForeignKey("organization.id"), nullable=False)
    org = db.relationship("Organization", backref="flowspec6")
    rstate_id = db.Column(db.Integer, db.ForeignKey("rstate.id"), nullable=False)
    rstate = db.relationship("Rstate", backref="flowspec6")

    def __init__(
        self,
        source,
        source_mask,
        source_port,
        destination,
        destination_mask,
        destination_port,
        next_header,
        flags,
        packet_len,
        expires,
        user_id,
        org_id,
        action_id,
        created=None,
        comment=None,
        rstate_id=1,
    ):
        self.source = source
        self.source_mask = source_mask
        self.dest = destination
        self.dest_mask = destination_mask
        self.source_port = source_port
        self.dest_port = destination_port
        self.next_header = next_header
        self.flags = flags
        self.packet_len = packet_len
        self.comment = comment
        self.expires = expires
        self.user_id = user_id
        self.org_id = org_id
        self.action_id = action_id
        if created is None:
            created = datetime.now()
        self.created = created
        self.rstate_id = rstate_id

    def __eq__(self, other):
        """
        Two models are equal if all the network parameters equals. User_id and time fields can differ.
        :param other: other Flowspec4 instance
        :return: boolean
        """
        return (
            self.source == other.source
            and self.source_mask == other.source_mask
            and self.dest == other.dest
            and self.dest_mask == other.dest_mask
            and self.source_port == other.source_port
            and self.dest_port == other.dest_port
            and self.next_header == other.next_header
            and self.flags == other.flags
            and self.packet_len == other.packet_len
            and self.action_id == other.action_id
            and self.rstate_id == other.rstate_id
        )

    def to_dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :returns: dictionary
        """
        if prefered_format == "timestamp":
            expires = int(datetime.timestamp(self.expires))
            created = int(datetime.timestamp(self.created))
        else:
            expires = utils.datetime_to_webpicker(self.expires, prefered_format)
            created = utils.datetime_to_webpicker(self.created, prefered_format)

        return {
            "id": str(self.id),
            "source": self.source,
            "source_mask": self.source_mask,
            "source_port": self.source_port,
            "dest": self.dest,
            "dest_mask": self.dest_mask,
            "dest_port": self.dest_port,
            "next_header": self.next_header,
            "flags": self.flags,
            "packet_len": self.packet_len,
            "comment": self.comment,
            "expires": expires,
            "created": created,
            "action": self.action.name,
            "user": self.user.uuid,
            "rstate": self.rstate.description,
        }

    def dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :returns: dictionary
        """
        return self.to_dict(prefered_format)

    def json(self, prefered_format="yearfirst"):
        """
        Serialize to json
        :param prefered_format: string with prefered time format
        :returns: json
        """
        return json.dumps(self.to_dict())



"""Integration tests for rule_service module."""

import pytest
from datetime import datetime, timedelta
from unittest.mock import patch

from flowapp import db, create_app
from flowapp.constants import RuleTypes
from flowapp.models import (
    RTBH,
    Flowspec4,
    Flowspec6,
    Whitelist,
    User,
    Organization,
    Role,
    Community,
    Action,
    Rstate,
)
from flowapp.services import rule_service


@pytest.fixture(scope="module")
def app():
    """Create and configure a Flask app for testing."""
    app = create_app("testing")

    # Create a test context
    with app.app_context():
        # Create database tables
        db.create_all()

        # Create test data
        _create_test_data()

        yield app

        # Clean up after tests
        db.session.remove()
        db.drop_all()


@pytest.fixture(scope="function")
def app_context(app):
    """Provide an application context for each test."""
    with app.app_context():
        yield


@pytest.fixture(scope="function")
def mock_announce_route():
    """Mock the announce_route function."""
    with patch("flowapp.services.rule_service.announce_route") as mock:
        yield mock


def _create_test_data():
    """Create test data in the database."""
    # Create roles
    admin_role = Role(name="admin", description="Administrator")
    user_role = Role(name="user", description="Normal user")
    view_role = Role(name="view", description="View only")
    db.session.add_all([admin_role, user_role, view_role])
    db.session.flush()

    # Create organizations
    org1 = Organization(name="Test Org 1", arange="192.168.1.0/24\n2001:db8::/64")
    org2 = Organization(name="Test Org 2", arange="10.0.0.0/8")
    db.session.add_all([org1, org2])
    db.session.flush()

    # Create users
    user1 = User(uuid="user1@example.com", name="User One")
    user1.role.append(user_role)
    user1.organization.append(org1)

    admin1 = User(uuid="admin@example.com", name="Admin User")
    admin1.role.append(admin_role)
    admin1.organization.append(org2)

    db.session.add_all([user1, admin1])
    db.session.flush()

    # Create rule states
    state_active = Rstate(description="active rule")
    state_withdrawn = Rstate(description="withdrawed rule")
    state_deleted = Rstate(description="deleted rule")
    state_whitelisted = Rstate(description="whitelisted rule")
    db.session.add_all([state_active, state_withdrawn, state_deleted, state_whitelisted])
    db.session.flush()

    # Create actions
    action1 = Action(name="Discard", command="discard", description="Drop traffic", role_id=user_role.id)
    action2 = Action(name="Rate limit", command="rate-limit 1000000", description="Limit traffic", role_id=user_role.id)
    db.session.add_all([action1, action2])
    db.session.flush()

    # Create communities
    community1 = Community(
        name="65000:1",
        comm="65000:1",
        larcomm="",
        extcomm="",
        description="Test community",
        role_id=user_role.id,
        as_path=False,
    )
    db.session.add(community1)
    db.session.flush()

    # Create some rules
    # IPv4 rule
    ipv4_rule = Flowspec4(
        source="192.168.1.1",
        source_mask=32,
        source_port="",
        dest="192.168.2.1",
        dest_mask=32,
        dest_port="",
        protocol="tcp",
        flags="",
        packet_len="",
        fragment="",
        expires=datetime.now() + timedelta(hours=1),
        comment="Test IPv4 rule",
        action_id=action1.id,
        user_id=user1.id,
        org_id=org1.id,
        rstate_id=state_active.id,
    )
    db.session.add(ipv4_rule)

    # IPv6 rule
    ipv6_rule = Flowspec6(
        source="2001:db8::1",
        source_mask=128,
        source_port="",
        dest="2001:db8:1::1",
        dest_mask=128,
        dest_port="",
        next_header="tcp",
        flags="",
        packet_len="",
        expires=datetime.now() + timedelta(hours=1),
        comment="Test IPv6 rule",
        action_id=action1.id,
        user_id=user1.id,
        org_id=org1.id,
        rstate_id=state_active.id,
    )
    db.session.add(ipv6_rule)

    # RTBH rule
    rtbh_rule = RTBH(
        ipv4="192.168.1.100",
        ipv4_mask=32,
        ipv6=None,
        ipv6_mask=None,
        community_id=community1.id,
        expires=datetime.now() + timedelta(hours=1),
        comment="Test RTBH rule",
        user_id=user1.id,
        org_id=org1.id,
        rstate_id=state_active.id,
    )
    db.session.add(rtbh_rule)

    # Save IDs as class variables for tests to use
    _create_test_data.user_id = user1.id
    _create_test_data.admin_id = admin1.id
    _create_test_data.org1_id = org1.id
    _create_test_data.org2_id = org2.id
    _create_test_data.ipv4_rule_id = ipv4_rule.id
    _create_test_data.ipv6_rule_id = ipv6_rule.id
    _create_test_data.rtbh_rule_id = rtbh_rule.id
    _create_test_data.action_id = action1.id
    _create_test_data.community_id = community1.id

    db.session.commit()


class TestRuleServiceIntegration:
    """Integration tests for rule_service functions."""

    def test_reactivate_rule_ipv4(self, app_context, mock_announce_route):
        """Test reactivating an IPv4 rule."""
        # Test data
        rule_id = _create_test_data.ipv4_rule_id
        user_id = _create_test_data.user_id
        org_id = _create_test_data.org1_id

        # Set new expiration time (2 hours in the future)
        new_expires = datetime.now() + timedelta(hours=2)
        new_comment = "Updated test comment"

        # Call the function
        model, messages = rule_service.reactivate_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=rule_id,
            expires=new_expires,
            comment=new_comment,
            user_id=user_id,
            org_id=org_id,
            user_email="test@example.com",
            org_name="Test Org",
        )

        # Verify the rule was updated
        assert model is not None
        assert model.id == rule_id
        assert model.comment == new_comment
        assert model.expires == new_expires
        assert model.rstate_id == 1  # active state

        # Verify route announcement was attempted
        assert mock_announce_route.called

        # Verify message
        assert messages == ["Rule successfully updated"]

    def test_reactivate_rule_inactive(self, app_context, mock_announce_route):
        """Test reactivating a rule to inactive state."""
        # Test data
        rule_id = _create_test_data.ipv4_rule_id
        user_id = _create_test_data.user_id
        org_id = _create_test_data.org1_id

        # Set past expiration time
        past_expires = datetime.now() - timedelta(hours=1)

        # Call the function
        model, messages = rule_service.reactivate_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=rule_id,
            expires=past_expires,
            comment="Expired comment",
            user_id=user_id,
            org_id=org_id,
            user_email="test@example.com",
            org_name="Test Org",
        )

        # Verify the rule was updated
        assert model is not None
        assert model.id == rule_id
        assert model.expires == past_expires
        assert model.rstate_id == 2  # inactive/withdrawn state

        # Verify route announcement was attempted
        assert mock_announce_route.called

    def test_delete_rule(self, app_context, mock_announce_route):
        """Test deleting a rule."""
        # Test data
        rule_id = _create_test_data.ipv6_rule_id
        user_id = _create_test_data.user_id

        # Call the function
        success, message = rule_service.delete_rule(
            rule_type=RuleTypes.IPv6,
            rule_id=rule_id,
            user_id=user_id,
            user_email="test@example.com",
            org_name="Test Org",
        )

        # Verify the rule was deleted
        assert success is True
        assert message == "Rule deleted successfully"

        # Verify the rule no longer exists in the database
        rule = db.session.get(Flowspec6, rule_id)
        assert rule is None

        # Verify route withdrawal was attempted
        assert mock_announce_route.called

    def test_delete_rule_not_found(self, app_context):
        """Test deleting a non-existent rule."""
        # Use a non-existent rule ID
        non_existent_id = 9999

        # Call the function
        success, message = rule_service.delete_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=non_existent_id,
            user_id=_create_test_data.user_id,
            user_email="test@example.com",
            org_name="Test Org",
        )

        # Verify the operation failed
        assert success is False
        assert message == "Rule not found"

    @patch("flowapp.services.rule_service.create_or_update_whitelist")
    def test_delete_rtbh_and_create_whitelist(self, mock_create_whitelist, app_context, mock_announce_route):
        """Test deleting an RTBH rule and creating a whitelist entry."""
        # Test data
        rule_id = _create_test_data.rtbh_rule_id
        user_id = _create_test_data.user_id
        org_id = _create_test_data.org1_id

        # Mock whitelist creation
        mock_whitelist = Whitelist(
            ip="192.168.1.100", mask=32, expires=datetime.now() + timedelta(days=7), user_id=user_id, org_id=org_id
        )
        mock_create_whitelist.return_value = (mock_whitelist, ["Whitelist created"])

        # Call the function
        success, messages, whitelist = rule_service.delete_rtbh_and_create_whitelist(
            rule_id=rule_id, user_id=user_id, org_id=org_id, user_email="test@example.com", org_name="Test Org"
        )

        # Verify success
        assert success is True
        assert len(messages) == 2
        assert "Rule deleted successfully" in messages[0]
        assert "Whitelist created" in messages[1]

        # Verify the rule was deleted
        rule = db.session.get(RTBH, rule_id)
        assert rule is None

        # Verify create_or_update_whitelist was called with correct data
        mock_create_whitelist.assert_called_once()
        args, kwargs = mock_create_whitelist.call_args
        form_data = kwargs.get("form_data", args[0] if args else None)
        assert form_data["ip"] == "192.168.1.100"
        assert form_data["mask"] == 32
        assert "Created from RTBH rule" in form_data["comment"]



from flask import Blueprint, jsonify


api = Blueprint("api_v2", __name__, template_folder="templates")
METHODS = ["GET", "POST", "PUT", "DELETE"]

@api.route("/", defaults={"path": ""}, methods=METHODS)
@api.route("/<path:path>", methods=METHODS)
def deprecated_warning(path):
    """Catch all routes and return a deprecated warning message."""
    message = "Warning: This API version is deprecated. Please use /api/v3/ instead."
    return jsonify({"message": message}), 400
